/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*-
 *
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Header: /nfs/jade/vint/CVSROOT/ns-2/mac/mac-802_11.cc,v 1.47 2004/04/02 01:00:25 xuanc Exp $
 *
 * Ported from CMU/Monarch's code, nov'98 -Padma.
 * Contributions by:
 *   - Mike Holland
 *   - Sushmita
 */

#include "delay.h"
#include "connector.h"
#include "packet.h"
#include "random.h"
#include "time.h"
#include "mobilenode.h"

// #define DEBUG 99

#include "arp.h"
#include "ll.h"
#include "mac.h"
#include "mac-timers.h"
#include "mac-802_11.h"
#include "mac-802_11-mlme.h"  // PSM stuff
#include "cmu-trace.h"
#include "span/hdr_span.h"

#define max(x,y) ((x > y) ? (x) : (y))

// Added by Sushmita to support event tracing
#include "agent.h"
#include "basetrace.h"

int hdr_beacon::offset_;
static class BeaconHeaderClass:public PacketHeaderClass { public:
  BeaconHeaderClass():PacketHeaderClass("PacketHeader/Beacon",
					sizeof(struct hdr_beacon)) {
    bind_offset(&hdr_beacon::offset_);
}} class_beaconhdr;


u_int32_t
PHY_MIB::getBeaconLen(struct hdr_beacon *bh) {
	return(getPLCPhdrLen() + sizeof(struct hdr_mac802_11)
	  // have to adjust for variable-length fields
	  + sizeof(struct beacon_body)
	  - sizeof(struct ssid) + ssid_size(&bh->bh_bb.bb_ssid)
	  - sizeof(struct supported_rates)
	  + supported_rates_size(&bh->bh_bb.bb_sr)
	  - sizeof(struct tim) + tim_size(&bh->bh_bb.bb_tim)
	  + ETHER_FCS_LEN);
}

/* our backoff timer doesn't count down in idle times during a
 * frame-exchange sequence as the mac tx state isn't idle; genreally
 * these idle times are less than DIFS and won't contribute to
 * counting down the backoff period, but this could be a real
 * problem if the frame exchange ends up in a timeout! in that case,
 * i.e. if a timeout happens we've not been counting down for the
 * duration of the timeout, and in fact begin counting down only
 * DIFS after the timeout!! we lose the timeout interval - which
 * will is not the REAL case! also, the backoff timer could be NULL
 * and we could have a pending transmission which we could have
 * sent! one could argue this is an implementation artifact which
 * doesn't violate the spec.. and the timeout interval is expected
 * to be less than DIFS .. which means its not a lot of time we
 * lose.. anyway if everyone hears everyone the only reason a ack will
 * be delayed will be due to a collision => the medium won't really be
 * idle for a DIFS for this to really matter!!
 */

inline void
Mac802_11::checkBackoffTimer()
{
	if(is_idle() && mhBackoff_.paused())
		mhBackoff_.resume(phymib_.getDIFS());
	if(! is_idle() && mhBackoff_.busy() && ! mhBackoff_.paused())
		mhBackoff_.pause();
}

inline void
Mac802_11::transmit(Packet *p, double timeout)
{
	struct hdr_mac802_11 *dh =
             (struct hdr_mac802_11*) hdr_mac::access(p);
	struct hdr_beacon *bh;
	tx_active_ = 1;

	if (txtime(p) != timeout) {
        //  printf("Timeouts do not match!\n");
        }
		
	if (EOTtarget_) {
		assert (eotPacket_ == NULL);
		eotPacket_ = p->copy();
	}

	/*
	 * If I'm transmitting without doing CS, such as when
	 * sending an ACK, any incoming packet will be "missed"
	 * and hence, must be discarded.
	 */
	if(rx_state_ != MAC_IDLE) {
		assert((dh->dh_fc.fc_type == MAC_Type_Control) ||
		       (dh->dh_fc.fc_type == MAC_Type_Management));
		assert(dh->dh_fc.fc_subtype == MAC_Subtype_ACK);
		assert(pktRx_);
		struct hdr_cmn *ch = HDR_CMN(pktRx_);
		ch->error() = 1;        /* force packet discard */
	}

	/*
	 * pass the packet on the "interface" which will in turn
	 * place the packet on the channel.
	 *
	 * NOTE: a handler is passed along so that the Network
	 *       Interface can distinguish between incoming and
	 *       outgoing packets.
	 */
	if ((dh->dh_fc.fc_type == MAC_Type_Management) &&
	    (dh->dh_fc.fc_subtype == MAC_Subtype_Beacon)) {
	    // if we are dealing with a beacon we need to fill in the tsf
	    // just before the first bit enters the PHY (c.f. 11.1.2)
	    bh = hdr_beacon::access(p);
	    bh->bh_bb.bb_ts = tsfTimer_.getTSF();
	}

	if (netif_->state() == pwr_state_sleep) {
#ifdef SPAN_DEBUG
		fprintf(stderr, "%f: _%d_ try to send, turn on radio\n",
		  Scheduler::instance().clock(), index_);
#endif
		netif_->set_state(pwr_state_idle);
	}

	downtarget_->recv(p->copy(), this);
	assert(!mhSend_.busy());
        if ( mhIF_.busy() ) {
		printf("Uh-oh, mhIF still busy! Check txtime values!\n");
        }
	assert(!mhIF_.busy());
	mhSend_.start(timeout);
	mhIF_.start(txtime(p));
}

inline void
Mac802_11::setRxState(MacState newState)
{
	rx_state_ = newState;
	checkBackoffTimer();
}

inline void
Mac802_11::setTxState(MacState newState)
{
	tx_state_ = newState;
	checkBackoffTimer();
}

void
Mac802_11::goto_sleep()
{
  struct buffer_entry *this_entry;
  int n = 0;
  CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
    n++;
  }
  
  if (n < 5) 
    netif_->set_state(pwr_state_sleep);
  else {
#if SPAN_DEBUG
    fprintf(stderr, "%f: _%d_ radio off denied, too many packets to send...\n",
	    Scheduler::instance().clock(), index_);
#endif
  }
}

void
Mac802_11::callback()
{
// This method tells the parent layer who pushed a packet into the 
// mac802_11 layer that it's finished sending the previous packet and is 
// ready for another.  This is how a parent queue gets drained.

  if (callback_) {
#ifdef SPAN_DEBUG
    fprintf(stderr, "%f: %d callback\n",
	    Scheduler::instance().clock(), index_);
#endif
    Handler *up = callback_;
    callback_ = 0;
    up->handle((Event *) 0);
  }
}


/* ======================================================================
   TCL Hooks for the simulator
   ====================================================================== */
static class Mac802_11Class : public TclClass {
public:
	Mac802_11Class() : TclClass("Mac/802_11") {
	  // srand((int) time(0));
	  // srandom(int) time(0));
	}
	TclObject* create(int, const char *const *) {
	  return (new Mac802_11());
	}
} class_mac802_11;


/* ======================================================================
   Mac  and Phy MIB Class Functions
   ====================================================================== */

PHY_MIB::PHY_MIB(Mac802_11 *parent)
{
	/*
	 * Bind the phy mib objects.  Note that these will be bound
	 * to Mac/802_11 variables
	 */

	parent->bind("CWMin_", &CWMin);
	parent->bind("CWMax_", &CWMax);
	parent->bind("SlotTime_", &SlotTime);
	parent->bind("SIFS_", &SIFSTime);
	parent->bind("PreambleLength_", &PreambleLength);
	parent->bind("PLCPHeaderLength_", &PLCPHeaderLength);
	parent->bind_bw("PLCPDataRate_", &PLCPDataRate);
}

MAC_MIB::MAC_MIB(Mac802_11 *parent)
{
	/*
	 * Bind the phy mib objects.  Note that these will be bound
	 * to Mac/802_11 variables
	 */
	
	parent->bind("RTSThreshold_", &RTSThreshold);
	parent->bind("ShortRetryLimit_", &ShortRetryLimit);
	parent->bind("LongRetryLimit_", &LongRetryLimit);

	// MaxTransmitMSDULifetime = MAC_MaxTransmitMSDULifetime;

}

/* ======================================================================
   Mac Class Functions
   ====================================================================== */
Mac802_11::Mac802_11() : 
	Mac(), phymib_(this), macmib_(this), mhIF_(this), mhNav_(this), 
	mhRecv_(this), mhSend_(this), 
	mhDefer_(this), mhBackoff_(this),
	mhTBTT_(this), mhBeacon_(this), mhAdvEnd_(this),
	mhEndATIM_(this), tsfTimer_(index_)
{
	nav_ = 0.0;
	tx_state_ = rx_state_ = MAC_IDLE;
	tx_active_ = 0;
	eotPacket_ = NULL;
	pktRTS_ = 0;
	pktCTRL_ = 0;		
	pktBEACON_ = 0;
	pktATIM_ = 0;

	cw_ = phymib_.getCWMin();
	ssrc_ = slrc_ = 0;
	atimrc_ = rtsrc_ = 0;

	// Added by Sushmita
        et_ = new EventTrace();
	
	sta_seqno_ = 1;
	cache_ = 0;
	cache_node_count_ = 0;

	initial_wake_count_ = 30;
	got_beacon_ = false;
	psm_mode_ = false;
	span_opt_ = false;

	macmib_.BeaconPeriod = 0.200;    // beacon period
	macmib_.ATIMWindow   = 0.040;    // atim window
	macmib_.AdvWindow    = 0.200;    // adv. traffic window

	// debug only
	entry_ = 0;

	CIRCLEQ_INIT(&buf_head_);

	total_retransmit = 0;
	total_wasted_time = 0;
	idle_start = 0.0;
	total_true_idle_time = 0.0;

	// chk if basic/data rates are set
	// otherwise use bandwidth_ as default;	
	Tcl& tcl = Tcl::instance();
	tcl.evalf("Mac/802_11 set basicRate_");
	if (strcmp(tcl.result(), "0") != 0) 
		bind_bw("basicRate_", &basicRate_);
	else
		basicRate_ = bandwidth_;

	tcl.evalf("Mac/802_11 set dataRate_");
	if (strcmp(tcl.result(), "0") != 0) 
		bind_bw("dataRate_", &dataRate_);
	else
		dataRate_ = bandwidth_;

        EOTtarget_ = 0;
       	bss_id_ = IBSS_ID;
	//printf("bssid in constructor %d\n",bss_id_);
}

void Mac802_11::Terminate()
{
  struct buffer_entry *this_entry;
  CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
    CIRCLEQ_REMOVE(&buf_head_, this_entry, entries);
    Packet *p = this_entry->packet_;
    drop(p, DROP_END_OF_SIMULATION);
    if (this_entry == entryATIM_) entryATIM_ = 0;
    delete this_entry;
  }
}

int
Mac802_11::command(int argc, const char*const* argv)
{
  if (argc == 2) {
     if (strcasecmp(argv[1], "reset") == 0) {
         Terminate();
         return TCL_OK;
     }
     if (strcasecmp(argv[1], "initialize") == 0) {
         // init();
         return TCL_OK;
     }
     if (strcasecmp(argv[1], "start") == 0) {
      // start generating beacons
      psm_mode_ = true;
      mhTBTT_.start(2 * macmib_.BeaconPeriod - fmod(tsfTimer_.getTSF(),
                    macmib_.BeaconPeriod * 1e6) / 1e6);
      return TCL_OK;
    }
    if (strcasecmp(argv[1], "spanopt") == 0) {
      span_opt_ = true;
      macmib_.BeaconPeriod = 0.300;    // beacon period
      macmib_.ATIMWindow   = 0.020;    // atim window
      macmib_.AdvWindow    = 0.100;    // adv. traffic window
      return TCL_OK;
    }

  } else if (argc == 3) {
     if (strcmp(argv[1], "eot-target") == 0) {
         EOTtarget_ = (NsObject*) TclObject::lookup(argv[2]);
         if (EOTtarget_ == 0)
            return TCL_ERROR;
         return TCL_OK;

     } else if (strcmp(argv[1], "bss_id") == 0) {
         bss_id_ = atoi(argv[2]);
         return TCL_OK;

     } else if (strcmp(argv[1], "log-target") == 0) { 
         logtarget_ = (NsObject*) TclObject::lookup(argv[2]);
         if(logtarget_ == 0)
            return TCL_ERROR;
         return TCL_OK;

     } else if(strcmp(argv[1], "nodes") == 0) {
         if(cache_) return TCL_ERROR;
         cache_node_count_ = atoi(argv[2]);
         cache_ = new Host[cache_node_count_ + 1];
         assert(cache_);
         bzero(cache_, sizeof(Host) * (cache_node_count_+1 ));
         return TCL_OK;

     } else if(strcmp(argv[1], "eventtrace") == 0) {
         // command added to support event tracing by Sushmita
         et_ = (EventTrace *)TclObject::lookup(argv[2]);
         return (TCL_OK);

     } else if(strcmp(argv[1], "setnode") == 0) {
         node_ = (MobileNode *) TclObject::lookup(argv[2]);
         if (!node_)  return TCL_ERROR;
         return TCL_OK;
     }
  }
  return Mac::command(argc, argv);
}

// Added by Sushmita to support event tracing
void Mac802_11::trace_event(char *eventtype, Packet *p) 
{
        if (et_ == NULL) return;
        char *wrk = et_->buffer();
        char *nwrk = et_->nbuffer();
	
        hdr_ip *iph = hdr_ip::access(p);
        //char *src_nodeaddr =
	//       Address::instance().print_nodeaddr(iph->saddr());
        //char *dst_nodeaddr =
        //      Address::instance().print_nodeaddr(iph->daddr());
	
        struct hdr_mac802_11* dh = HDR_MAC802_11(p);
	
        //struct hdr_cmn *ch = HDR_CMN(p);
	
	if(wrk != 0) {
		sprintf(wrk, "E -t "TIME_FORMAT" %s %2x ",
			et_->round(Scheduler::instance().clock()),
                        eventtype,
                        //ETHER_ADDR(dh->dh_sa)
                        ETHER_ADDR(dh->dh_ta)
                        );
        }
        if(nwrk != 0) {
                sprintf(nwrk, "E -t "TIME_FORMAT" %s %2x ",
                        et_->round(Scheduler::instance().clock()),
                        eventtype,
                        //ETHER_ADDR(dh->dh_sa)
                        ETHER_ADDR(dh->dh_ta)
                        );
        }
        et_->dump();
}

/* ======================================================================
   Debugging Routines
   ====================================================================== */

double Mac802_11::bftime()
{
  return mhBackoff_.total_bftime;
}

void
Mac802_11::trace_pkt(Packet *p) 
{
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac802_11* dh = HDR_MAC802_11(p);
	u_int16_t *t = (u_int16_t*) &dh->dh_fc;

	fprintf(stderr, "\t[ %2x %2x %2x %2x ] %x %s %d\n",
		*t, dh->dh_duration,
		 ETHER_ADDR(dh->dh_ra), ETHER_ADDR(dh->dh_ta),
		index_, packet_info.name(ch->ptype()), ch->size());
}

void
Mac802_11::dump(char *fname)
{
	fprintf(stderr,
		"\n%s --- (INDEX: %d, time: %2.9f)\n",
		fname, index_, Scheduler::instance().clock());

	fprintf(stderr,
		"\ttx_state_: %x, rx_state_: %x, nav: %2.9f, idle: %d\n",
		tx_state_, rx_state_, nav_, is_idle());

	fprintf(stderr,
		"\tpktTx_: %x, pktRx_: %x, pktRTS_: %x, pktCTRL_: %x, callback: %x\n",
		(int) pktTx_, (int) pktRx_, (int) pktRTS_,
		(int) pktCTRL_, (int) callback_);

	fprintf(stderr,
		"\tDefer: %d, Backoff: %d (%d), Recv: %d, Timer: %d Nav: %d\n",
		mhDefer_.busy(), mhBackoff_.busy(), mhBackoff_.paused(),
		mhRecv_.busy(), mhSend_.busy(), mhNav_.busy());
	fprintf(stderr,
		"\tBackoff Expire: %f\n",
		mhBackoff_.expire());
}


/* ======================================================================
   Packet Headers Routines
   ====================================================================== */
int
Mac802_11::hdr_dst(char* hdr, int dst )
{
	struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;
	
       if (dst > -2) {
               if ((bss_id() == IBSS_ID) || (addr() == bss_id())) {
                       /* if I'm AP (2nd condition above!), the dh_3a
                        * is already set by the MAC whilst fwding; if
                        * locally originated pkt, it might make sense
                        * to set the dh_3a to myself here! don't know
                        * how to distinguish between the two here - and
                        * the info is not critical to the dst station
                        * anyway!
                        */
                       STORE4BYTE(&dst, (dh->dh_ra));
               } else {
                       /* in BSS mode, the AP forwards everything;
                        * therefore, the real dest goes in the 3rd
                        * address, and the AP address goes in the
                        * destination address
                        */
                       STORE4BYTE(&bss_id_, (dh->dh_ra));
                       STORE4BYTE(&dst, (dh->dh_3a));
               }
       }


       return (u_int32_t)ETHER_ADDR(dh->dh_ra);
}

int 
Mac802_11::hdr_src(char* hdr, int src )
{
	struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;
        if(src > -2)
               STORE4BYTE(&src, (dh->dh_ta));
        return ETHER_ADDR(dh->dh_ta);
}

inline int 
Mac802_11::hdr_type(char* hdr, u_int16_t type)
{
	struct hdr_mac802_11 *dh = (struct hdr_mac802_11*) hdr;
	if(type)
		STORE2BYTE(&type,(dh->dh_body));
	return GET2BYTE(dh->dh_body);
}


/* ======================================================================
   Misc Routines
   ====================================================================== */

Packet * Mac802_11::reclaim_packet(int dst)
{ 
  struct buffer_entry *this_entry;
  CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) { 
    if (this_entry->packet_ && this_entry->addr(this) == (unsigned)dst) {
      CIRCLEQ_REMOVE(&buf_head_, this_entry, entries); 
      Packet *p = this_entry->packet_;
      if (this_entry == entryATIM_) entryATIM_ = 0;
      delete this_entry;
      return p;
    }
  }
  return 0;
}

inline int
Mac802_11::is_idle()
{
	if(rx_state_ != MAC_IDLE)
		return 0;
	if(tx_state_ != MAC_IDLE)
		return 0;
	if(nav_ > Scheduler::instance().clock())
		return 0;
	
	return 1;
}

void
Mac802_11::discard(Packet *p, const char* why)
{
	hdr_mac802_11* mh = HDR_MAC802_11(p);
	hdr_cmn *ch = HDR_CMN(p);

	/* if the rcvd pkt contains errors, a real MAC layer couldn't
	   necessarily read any data from it, so we just toss it now */
	if(ch->error() != 0) {
		Packet::free(p);
		return;
	}

	switch(mh->dh_fc.fc_type) {
	case MAC_Type_Management:
		if((u_int32_t)ETHER_ADDR(mh->dh_ra) == \
		  (u_int32_t)index_ ||
		  (u_int32_t)ETHER_ADDR(mh->dh_ta) == \
		  (u_int32_t)index_ ||
		  (u_int32_t)ETHER_ADDR(mh->dh_ra) == MAC_BROADCAST) {
			drop(p,why);
			return;
		}
		break;
	case MAC_Type_Control:
		switch(mh->dh_fc.fc_subtype) {
		case MAC_Subtype_RTS:
			 if((u_int32_t)ETHER_ADDR(mh->dh_ta) ==  (u_int32_t)index_) {
				drop(p, why);
				return;
			}
			/* fall through - if necessary */
		case MAC_Subtype_CTS:
		case MAC_Subtype_ACK:
			if((u_int32_t)ETHER_ADDR(mh->dh_ra) == (u_int32_t)index_) {
				drop(p, why);
				return;
			}
			break;
		default:
			fprintf(stderr, "invalid MAC Control subtype\n");
			exit(1);
		}
		break;
	case MAC_Type_Data:
		switch(mh->dh_fc.fc_subtype) {
		case MAC_Subtype_Data:
			if((u_int32_t)ETHER_ADDR(mh->dh_ra) == \
                           (u_int32_t)index_ ||
                          (u_int32_t)ETHER_ADDR(mh->dh_ta) == \
                           (u_int32_t)index_ ||
                          (u_int32_t)ETHER_ADDR(mh->dh_ra) == MAC_BROADCAST) {
                                drop(p,why);
                                return;
			}
			break;
		default:
			fprintf(stderr, "invalid MAC Data subtype\n");
			exit(1);
		}
		break;
	default:
		fprintf(stderr, "invalid MAC type (%x)\n", mh->dh_fc.fc_type);
		trace_pkt(p);
		exit(1);
	}
	Packet::free(p);
}

void
Mac802_11::capture(Packet *p)
{
	/*
	 * Update the NAV so that this does not screw
	 * up carrier sense.
	 */	
	set_nav(usec(phymib_.getEIFS() + txtime(p)));
	Packet::free(p);
}

void
Mac802_11::collision(Packet *p)
{
	switch(rx_state_) {
	case MAC_RECV:
		setRxState(MAC_COLL);
		/* fall through */
	case MAC_COLL:
		assert(pktRx_);
		assert(mhRecv_.busy());
		/*
		 *  Since a collision has occurred, figure out
		 *  which packet that caused the collision will
		 *  "last" the longest.  Make this packet,
		 *  pktRx_ and reset the Recv Timer if necessary.
		 */
		if(txtime(p) > mhRecv_.expire()) {
			mhRecv_.stop();
			discard(pktRx_, DROP_MAC_COLLISION);
			pktRx_ = p;
			mhRecv_.start(txtime(pktRx_));
		}
		else {
			discard(p, DROP_MAC_COLLISION);
		}
		break;
	default:
		assert(0);
	}
}

void
Mac802_11::tx_resume()
{
	double rTime;
	bool active = false; 
	assert(mhSend_.busy() == 0);

	if (mhEndATIM_.busy())
	  // Case where we're inside the ATIM window

          if (!got_beacon_) {
          // inside atim window, no beacon heard, but has a
          // control packet, need to send it.
            if ((pktCTRL_) && (mhDefer_.busy()==0))
              // mhDefer.start(sifs, "sifs defer for cts/ack");
	      mhDefer_.start(phymib_.getSIFS());
          } else { 
	    // We have heard the beacon already for this ATIM window
	    if (pktCTRL_) {
	      if (mhDefer_.busy() == 0)
	    //     mhDefer_.start(sifs, "sifs defer for cts/ack");
	           mhDefer_.start(phymib_.getSIFS());
	      active = true;
	    }

	    if (!active) {
	      atim_scan();
	      if (pktBEACON_ || pktATIM_) {
	        if ((mhBackoff_.busy() == 0) && (mhDefer_.busy() == 0)) {
	          rTime = (Random::random() % cw_) * 
	                   phymib_.getSlotTime();
	          mhDefer_.start( phymib_.getDIFS() + rTime);
 	          // mhDefer_.start(difs);
	        }  
	      } else {
	        // Just normal (non-ATIM) packet
	        callback();
	    }
	  }

	} else {
	  // Case where we're outside the ATIM window

	  // assert(mhDefer_.busy() == 0);
	  if(pktCTRL_) {
	     if (mhDefer_.busy()==0)  // need to send a CTS or ACK
		mhDefer_.start(phymib_.getSIFS());
	     //  mhDefer_.start(sifs, "sifs defer for cts/ack");
	     active = true;
	  }

	  else if(pktRTS_) {
	     if ((mhBackoff_.busy() == 0) && (mhDefer_.busy() == 0)) {
	        rTime = (Random::random() % cw_) * phymib_.getSlotTime();
	        mhDefer_.start( phymib_.getDIFS() + rTime);
	     }
	     active = true;
	  }

	  else if(pktTx_) {
	     if ((mhBackoff_.busy() == 0) && (mhDefer_.busy() == 0)) {
	        hdr_cmn *ch = HDR_CMN(pktTx_);
                struct hdr_mac802_11 *mh = HDR_MAC802_11(pktTx_);
			
	        if ((u_int32_t) ch->size() < macmib_.getRTSThreshold() ||
	         (u_int32_t) ETHER_ADDR(mh->dh_ra) == MAC_BROADCAST) {
		  rTime = (Random::random() % cw_)
	                 * phymib_.getSlotTime();
	          mhDefer_.start(phymib_.getDIFS() + rTime);
                } else {
	          mhDefer_.start(phymib_.getSIFS());
                }
	     }
	     active = true;
	  }

	  if (!active) {
	    // Still in the non-ATIM window, but we didn't have any 
	    // immediate packets of Tx, RTS, or CTRL.  So now scan the 
	    // queue for packets.
	    tx_scan();

	    if (pktRTS_ || pktTx_) {
	      if (mhBackoff_.busy() == 0) {
               // If we are already deferring, no need to reset defer timer 
	        if (is_idle())
                  if (mhDefer_.busy()==0) {
	            rTime = (Random::random() % cw_)
	                 * phymib_.getSlotTime();
	            mhDefer_.start(phymib_.getDIFS() + rTime);
	          }
	        else // not IDLE, then start backoff timer
		  mhBackoff_.start(cw_, is_idle());
	      }

	    } else { // Not pktRTS or pktTx
	      has_packet_tx_ = 0;
	      // no data to send nor receive, so can go to sleep if needs to
	      if (!initial_wake_count_ &&
	         ((!recvd_bcast_atim_ && !recvd_ucast_atim_)
	          || !mhAdvEnd_.busy()) && netif_->PowerSave()) {
#ifdef SPAN_DEBUG
	        fprintf(stderr, "%f: _%d_ done tx, nothing to rx, ",
	            Scheduler::instance().clock(), index);
	        fprintf(stderr, "turn radio off?\n");
#endif
	        goto_sleep();
	      } else {
	        callback();
#if SPAN_DEBUG > 1
	        fprintf(stderr,
		  "%f: _%d_ done tx, stay on, b %d, u %d, w %d\n",
		  Scheduler::instance().clock(), index,
		  recvd_bcast_atim_, recvd_ucast_atim_,
		  initial_wake_count_);
#endif
	      }
            }
          }
	}
	setTxState(MAC_IDLE);
}

void
Mac802_11::rx_resume()
{
	assert(pktRx_ == 0);
	assert(mhRecv_.busy() == 0);
	setRxState(MAC_IDLE);

	if (span_opt_) {
	   if (netif_->PowerSave() && (netif_->state() == pwr_state_idle) &&
	      !mhAdvEnd_.busy() && !pktTx_ && !pktRTS_ &&
	      (rx_state_ == MAC_IDLE) && (tx_state_ == MAC_IDLE) && 
	      !initial_wake_count_) {
#ifdef SPAN_DEBUG
	      fprintf(stderr, "%f: _%d_ end of adv win, turn off radio?\n",
	      Scheduler::instance().clock(), index);
#endif
	      goto_sleep();
	   }
	}
}

bool Mac802_11::tx_scan()
{
  if (pktTx_)
    return false;
  bool remove = false;
  struct buffer_entry *this_entry = 0;

  // try to find packet with an ACK, or broadcast packets
  if (!remove && mhAdvEnd_.busy()) {
    CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
      assert(this_entry);
      u_int32_t dst = ETHER_ADDR( ((struct hdr_mac802_11*)
              hdr_mac::access(this_entry->packet_))->dh_ra);

      if (!this_entry->packet_ || !this_entry->pwr_mgt_ ||
	  (!this_entry->recvd_ack_ && dst != MAC_BROADCAST) ||
	  (dst == MAC_BROADCAST && !this_entry->sent_atim_))
	continue;

      this_entry->sent_data_ = true;
      sendDATA(this_entry->packet_);
      sendRTS(dst);

      CIRCLEQ_REMOVE(&buf_head_, this_entry, entries);
#if SPAN_DEBUG > 1
      fprintf(stderr, "%f: %d tx_scan sending advertised packet\n",
	      Scheduler::instance().clock(), index);
#endif
      remove = true;
      break;
    }
  }

  // try to find an un-advertised packet to a node not in pwr mgt mode
  if (!remove) {
    CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
      assert(this_entry);
      u_int32_t dst = ETHER_ADDR( ((struct hdr_mac802_11*)
              hdr_mac::access(this_entry->packet_))->dh_ra);

      // if unicast dst in pwr mgt mode but did not receive ack, or if
      // un-advertised broadcast traffic, skip
      if (!this_entry->packet_ ||
	  (this_entry->pwr_mgt_ && !this_entry->recvd_ack_
	   && dst != MAC_BROADCAST) ||
	  (dst == MAC_BROADCAST && !this_entry->sent_atim_))
	continue;

      // if dst is in power saving mode and advertised traffic window is
      // over, skip packet until next round.
      if (!mhAdvEnd_.busy() && this_entry->pwr_mgt_)
	continue;

      this_entry->sent_data_ = true;
      sendDATA(this_entry->packet_);
      sendRTS(dst);

      CIRCLEQ_REMOVE(&buf_head_, this_entry, entries);
#if SPAN_DEBUG > 1
      fprintf(stderr, "%f: %d tx_scan sending un-advertised packet\n",
	      Scheduler::instance().clock(), index);
#endif
      remove = true;
      break;
    }
  }

  if (remove) {
    if (this_entry == entryATIM_) entryATIM_ = 0;
    delete this_entry;
    return true;
  }
  return false;
}


void Mac802_11::atim_scan()
{
  // scan for an unsent ATIM in the buffer and send it.
  struct buffer_entry *this_entry;
  if (!pktATIM_) {
    CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
      if (!this_entry->sent_atim_ && !this_entry->recvd_ack_
	  && this_entry->pwr_mgt_) 
      {
	pktATIM_ = this_entry->packet_->copy();
	entryATIM_ = this_entry;

	// mac 802.11 header
	struct hdr_mac802_11 *dh = HDR_MAC802_11(pktATIM_);
	dh->dh_fc.fc_type = MAC_Type_Management;
	dh->dh_fc.fc_subtype = MAC_Subtype_ATIM;
	dh->dh_fc.fc_pwr_mgt = netif_->PowerSave()? 1 : 0;
        
	if (ETHER_ADDR(dh->dh_ra) != MAC_BROADCAST) 
	  dh->dh_duration = usec( txtime(phymib_.getACKlen(), basicRate_)
 		                  + phymib_.getSIFS() );
	else 
	  dh->dh_duration = 0;

	// common header
	struct hdr_cmn *ch = HDR_CMN(pktATIM_);
	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->iface() = -2;
	ch->error() = 0;
	ch->size() = phymib_.getHdrLen11();
 	ch->txtime() = txtime(ch->size(), basicRate_);
	break;
      }
    }
  }
}


/* ======================================================================
   Timer Handler Routines
   ====================================================================== */
void
Mac802_11::backoffHandler()
{

	if(pktCTRL_) {
	  assert(mhSend_.busy() || mhDefer_.busy());
	  return;
	}

	if (mhEndATIM_.busy()) {

	  // we are almost or inside the ATIM window
	  if (!got_beacon_) {
	    check_pktBEACON();
	    return;
	  }

	  atim_scan();
	  if (check_pktATIM() == 0) {
	     if (entryATIM_)  entryATIM_->sent_atim_ = true;
	     return;
	  }

	} else {
	  // Normal traffic time

	  if(check_pktRTS() == 0) return;
	  if(check_pktTx() == 0) return;

	  tx_scan();

	  if(check_pktRTS() == 0) return;
	  if(check_pktTx() == 0) return;
	  has_packet_tx_ = 0;
	}
}

void
Mac802_11::deferHandler()
{

	if (mhEndATIM_.busy()) { // We are inside the ATIM window
	  if (check_pktCTRL() == 0)  return;

	  if (got_beacon_) {
	     atim_scan();
	     if (check_pktATIM() == 0) {
	       if (entryATIM_) entryATIM_->sent_atim_ = true;
               return;
	     }
	  }

	} else {  // We are outside the ATIM window	
	  if(check_pktCTRL() == 0) return;
	  assert(mhBackoff_.busy() == 0);
	  if(check_pktRTS() == 0) return;
	  if(check_pktTx() == 0) return;
          tx_resume();
	}
}

void
Mac802_11::navHandler()
{
	if(is_idle() && mhBackoff_.paused())
		mhBackoff_.resume(phymib_.getDIFS());
}

void
Mac802_11::recvHandler()
{
	recv_timer();
}

void
Mac802_11::sendHandler()
{
	send_timer();
}


void
Mac802_11::txHandler()
{
	if (EOTtarget_) {
		assert(eotPacket_);
		EOTtarget_->recv(eotPacket_, (Handler *) 0);
		eotPacket_ = NULL;
	}
	tx_active_ = 0;
}

void Mac802_11::tbttHandler()

{
  // schedule end of advertised traffic window
  if (mhAdvEnd_.busy())  mhAdvEnd_.stop();
  mhAdvEnd_.start(macmib_.AdvWindow);

  // schedule the end of the ATIM window
  if (mhEndATIM_.busy())  mhEndATIM_.stop();
  mhEndATIM_.start(macmib_.ATIMWindow);

#if SPAN_DEBUG
  if (netif_->state() != pwr_state_sleep) {
    fprintf(stderr, "xx %f: _%d_ atim win, was %c\n",
	    Scheduler::instance().clock(), index,
	    (netif_->state() == pwr_state_sleep) ? '~' : '^');
  }
#endif
  // immediately wake up if not already awake
  if (netif_->state() == pwr_state_sleep) {
#if SPAN_DEBUG
    fprintf(stderr, "xx %f: _%d_ atim win, turning radio on\n",
	    Scheduler::instance().clock(), index);
#endif
    netif_->set_state(pwr_state_idle);
  }

  // our next TBTT is exactly one BeaconPeriod away
  double tbtt = macmib_.BeaconPeriod -
      fmod(tsfTimer_.getTSF(), macmib_.BeaconPeriod * 1e6) / 1e6;
  if (tbtt < macmib_.BeaconPeriod / 3)
    tbtt += macmib_.BeaconPeriod;
  assert(!mhTBTT_.busy());
  mhTBTT_.start(tbtt);
  assert(!mhBeacon_.busy());
  mhBeacon_.start(0.0001);

  got_beacon_ = false;
  first_atim_ = true;
  recvd_atim_ = false;
  has_packet_tx_ = 0;
  recvd_bcast_atim_ = 0;
  recvd_ucast_atim_ = 0;

  struct buffer_entry *this_entry;

  // expire "old" atims
  CIRCLEQ_HEAD(buffer_head, buffer_entry) expired;
  CIRCLEQ_INIT(&expired);

  CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
    if ((this_entry->age_ >= 2) || this_entry->sent_data_) {
      CIRCLEQ_REMOVE(&buf_head_, this_entry, entries);
      CIRCLEQ_INSERT_TAIL(&expired, this_entry, entries);
    }
    else {
      assert(this_entry);
#if SPAN_DEBUG > 1
      if (this_entry->sent_atim_) {
        unsigned dst = this_entry->addr(this);
        fprintf(stderr, "%f: %d recycle entry: pwr %d, ack %d, to %d\n",
		Scheduler::instance().clock(), index,
		this_entry->pwr_mgt_, this_entry->recvd_ack_, dst);
      }
#endif
      // retry ATIM transmission at the beginning of the ATIM window,
      // but don't reset pwr_mgt_, so we don't send extra ATIMs.
      this_entry->sent_atim_ = this_entry->recvd_ack_ = false;
      this_entry->pwr_mgt_updated_ = false;
      this_entry->age_++;
    }
  }

  CIRCLEQ_FOREACH(this_entry, &expired, entries) { 
    if (!this_entry->sent_data_) { 
      /* tell the callback the send operation failed 
       * before discarding the packet */ 
      hdr_cmn *ch = HDR_CMN(this_entry->packet_); 
      if (ch->xmit_failure_) { 
	ch->xmit_reason_ = XMIT_REASON_BUF; 
	ch->xmit_failure_ 
	  (this_entry->packet_->copy(), ch->xmit_failure_data_); 
      } 
      discard(this_entry->packet_, DROP_MAC_BUFFER_FULL); 
    } 
    if (this_entry == entryATIM_) entryATIM_ = 0;
    delete this_entry;
  }
}

void Mac802_11::beaconHandler()
{
  sendBEACON();
  if (tx_state_ == MAC_IDLE && !mhSend_.busy() &&
      !mhBackoff_.busy() && !mhDefer_.busy()) 
    // mhBackoff_.start(DSSS_CWMin * 2, 0.0, is_idle(), "beacon");
    mhBackoff_.start(DSSS_CWMin * 2, is_idle());
}

void Mac802_11::advEndHandler()
{
  if (span_opt_) {

    if (netif_->PowerSave() && (netif_->state() == pwr_state_idle) &&
        !pktTx_ && !pktRTS_ && (rx_state_ == MAC_IDLE) &&
        (tx_state_ == MAC_IDLE) && !initial_wake_count_) {
#ifdef SPAN_DEBUG
      fprintf(stderr, "%f: _%d_ end of adv win, turn off radio?\n",
	      Scheduler::instance().clock(), index);
#endif
      goto_sleep();
    }
  }
}

void Mac802_11::atimEndHandler()

{
  double atimPendingCorrection = 0.0;

  assert(tx_state_ != MAC_CTS);

  if (tx_state_ == MAC_ATIM)
    // if we are currently transmitting an ATIM when the ATIM window 
    // ends, add a correction to the backoff timer.
    atimPendingCorrection = max(mhSend_.expire(), mhIF_.expire()) + 0.000001;
  
  // end of the ATIM window -- search for acked buffered packets in 
  // the buffer and send data after a backoff...
  if (tx_scan() || pktTx_) {
    // we have data to send
    if (!mhDefer_.busy() && !mhBackoff_.busy() && !mhSend_.busy())
      mhBackoff_.start(cw_, 1);
  //   mhBackoff_.start(cw, atimPendingCorrection, 1, "end of atim");
#if SPAN_DEBUG > 1
    fprintf(stderr, "%f: _%d_ has pkt to send, tx %d, b %d, u %d\n",
	    Scheduler::instance().clock(), index,
	    has_packet_tx_, recvd_bcast_atim_, recvd_ucast_atim_);
#endif
  } 
  else if ((!recvd_atim_ && !initial_wake_count_ && !has_packet_tx_) && 
           (netif_->PowerSave())) {
#ifdef SPAN_DEBUG
    fprintf(stderr, "%f: _%d_ no atim, turn off radio?\n",
	    Scheduler::instance().clock(), index);
#endif
    goto_sleep();
  } else if (netif_->PowerSave()) {
#if SPAN_DEBUG > 1
    fprintf(stderr, "%f: _%d_ end of atim, stay on, tx %d, b %d, u %d\n",
	    Scheduler::instance().clock(), index,
	    has_packet_tx_, recvd_bcast_atim_, recvd_ucast_atim_);
#endif
  }

  if (initial_wake_count_)
    initial_wake_count_--;
}


/* ======================================================================
   The "real" Timer Handler Routines
   ====================================================================== */
void
Mac802_11::send_timer()
{
	switch(tx_state_) {
	/*
	 * Sent a RTS, but did not receive a CTS.
	 */
	case MAC_RTS:
		RetransmitRTS();
		break;
	/*
	 * Sent a CTS, but did not receive a DATA packet.
	 */
	case MAC_CTS:
		assert(pktCTRL_);
		Packet::free(pktCTRL_); 
		pktCTRL_ = 0;
		break;
	/*
	 * Sent DATA, but did not receive an ACK packet.
	 */
	case MAC_SEND:
		RetransmitDATA();
		break;
	/*
	 * Sent an ACK, and now ready to resume transmission.
	 */
	case MAC_ACK:
		assert(pktCTRL_);
		Packet::free(pktCTRL_); 
		pktCTRL_ = 0;
		break;

	/*
	 * Sent a Beacon, and now ready to resume transmission.
	 */
	case MAC_BEACON:
		if (pktBEACON_) {
		  Packet::free(pktBEACON_);
		  pktBEACON_ = 0;
		}
		break;

	/* 
	 * Sent an ATIM, but did not receive an ACK packet.  Could
	 * have been a multicast ATIM.  RetransmitATIM checks this
	 * case.
 	 */ 
	case MAC_ATIM:
		RetransmitATIM();
		break;

	case MAC_IDLE:
		break;
	default:
		assert(0);
	}
	tx_resume();
}

/* ======================================================================
   Outgoing Packet Routines
   ====================================================================== */
int
Mac802_11::check_pktCTRL()
{
	struct hdr_mac802_11 *mh;
	double timeout;

	if(pktCTRL_ == 0)
		return -1;
	if(tx_state_ == MAC_CTS || tx_state_ == MAC_ACK)
		return -1;

	mh = HDR_MAC802_11(pktCTRL_);
							  
	switch(mh->dh_fc.fc_subtype) {
	/*
	 *  If the medium is not IDLE, don't send the CTS.
	 */
	case MAC_Subtype_CTS:
		if(!is_idle()) {
			discard(pktCTRL_, DROP_MAC_BUSY); pktCTRL_ = 0;
			return 0;
		}
		setTxState(MAC_CTS);
		/*
		 * timeout:  cts + data tx time calculated by
		 *           adding cts tx time to the cts duration
		 *           minus ack tx time -- this timeout is
		 *           a guess since it is unspecified
		 *           (note: mh->dh_duration == cf->cf_duration)
		 */		
		 timeout = txtime(phymib_.getCTSlen(), basicRate_)
                        + DSSS_MaxPropagationDelay                      // XXX
                        + sec(mh->dh_duration)
                        + DSSS_MaxPropagationDelay                      // XXX
                       - phymib_.getSIFS()
                       - txtime(phymib_.getACKlen(), basicRate_);
		break;
		/*
		 * IEEE 802.11 specs, section 9.2.8
		 * Acknowledments are sent after an SIFS, without regard to
		 * the busy/idle state of the medium.
		 */
	case MAC_Subtype_ACK:		
		setTxState(MAC_ACK);
		timeout = txtime(phymib_.getACKlen(), basicRate_);
		break;
	default:
		fprintf(stderr, "check_pktCTRL:Invalid MAC Control subtype\n");
		exit(1);
	}
	transmit(pktCTRL_, timeout);
	return 0;
}

int
Mac802_11::check_pktRTS()
{
	struct hdr_mac802_11 *mh;
	double timeout;

	assert(mhBackoff_.busy() == 0);

	if(pktRTS_ == 0)
 		return -1;
	mh = HDR_MAC802_11(pktRTS_);

 	switch(mh->dh_fc.fc_subtype) {
	case MAC_Subtype_RTS:
		if(! is_idle()) {
			inc_cw();
			mhBackoff_.start(cw_, is_idle());
			return 0;
		}
		setTxState(MAC_RTS);
		timeout = txtime(phymib_.getRTSlen(), basicRate_)
			+ DSSS_MaxPropagationDelay                      // XXX
			+ phymib_.getSIFS()
			+ txtime(phymib_.getCTSlen(), basicRate_)
			+ DSSS_MaxPropagationDelay;
		break;
	default:
		fprintf(stderr, "check_pktRTS:Invalid MAC Control subtype\n");
		exit(1);
	}
	transmit(pktRTS_, timeout);
  

	return 0;
}

int
Mac802_11::check_pktTx()
{
	struct hdr_mac802_11 *mh;
	double timeout;
	
	assert(mhBackoff_.busy() == 0);

	if(pktTx_ == 0)
		return -1;

	mh = HDR_MAC802_11(pktTx_);

	switch(mh->dh_fc.fc_subtype) {
	case MAC_Subtype_Data:
		if(! is_idle()) {
			sendRTS(ETHER_ADDR(mh->dh_ra));
			inc_cw();
			mhBackoff_.start(cw_, is_idle());
			return 0;
		}
		setTxState(MAC_SEND);
		if((u_int32_t)ETHER_ADDR(mh->dh_ra) != MAC_BROADCAST)
                        timeout = txtime(pktTx_)
                                + DSSS_MaxPropagationDelay              // XXX
                               + phymib_.getSIFS()
                               + txtime(phymib_.getACKlen(), basicRate_)
                               + DSSS_MaxPropagationDelay;             // XXX
		else
			timeout = txtime(pktTx_);
		break;
	default:
		fprintf(stderr, "check_pktTx:Invalid MAC Control subtype\n");
		exit(1);
	}
	transmit(pktTx_, timeout);
	return 0;
}

int Mac802_11::check_pktBEACON()
{
  struct hdr_mac802_11 *mh;

  if (!pktBEACON_)  return -1;
  if (tx_state_ != MAC_IDLE)  return -1;
  mh = HDR_MAC802_11(pktBEACON_);
  switch (mh->dh_fc.fc_subtype) {
  case MAC_Subtype_Beacon:
    setTxState(MAC_BEACON);
    break;

  default:
    fprintf(stderr, "Invalid MAC Control subtype\n");
    exit(1);
  }

  got_beacon_ = true;
  transmit(pktBEACON_, txtime(pktBEACON_));
  return 0;
}

int Mac802_11::check_pktATIM()
{
  struct hdr_mac802_11 *atim_mh;
  struct hdr_cmn *ch;

  assert(!mhBackoff_.busy());
  if (tx_state_ != MAC_IDLE)  return -1;
  if (pktATIM_ == 0)  return -1;

  if (entryATIM_ == 0) {
    Packet::free(pktATIM_);
    pktATIM_ = 0;
    return -1;
  }

  atim_mh = HDR_MAC802_11(pktATIM_);
  ch = HDR_CMN(pktATIM_);
  if (!is_idle()) {
    if (!mhBackoff_.busy()) {
      inc_cw();
      mhBackoff_.start(cw_, is_idle());
    }
    return -1;
  } else {
    setTxState(MAC_ATIM);
    // create an ATIM packet based on pktATIM
    atim_mh->dh_fc.fc_type = MAC_Type_Management;
    atim_mh->dh_fc.fc_subtype = MAC_Subtype_ATIM;

    double timeout;
    if (ETHER_ADDR(atim_mh->dh_ra) != MAC_BROADCAST)
      // (DATA_Time(len) + ACK_Time + sifs + difs)
      timeout = ( txtime(ch->size(), basicRate_) + 
                  txtime(phymib_.getACKlen(), basicRate_) +
                  phymib_.getSIFS() + phymib_.getDIFS() );
    else
      timeout = txtime(pktATIM_);

    // store the address of the STA we last transmitted an ATIM
    // to
    int da = ETHER_ADDR(atim_mh->dh_ra);
    STORE4BYTE(&da, (last_atim_ra_));

    if (ch->next_hop_ == (int) MAC_BROADCAST)
      has_packet_tx_++;
#if SPAN_DEBUG > 1
    fprintf(stderr, "%f: _%d_ send ATIM %d, %d\n",
	    Scheduler::instance().clock(),
	    index, ch->next_hop_, has_packet_tx_);
#endif
    // transmit the ATIM
    transmit(pktATIM_, timeout);
  }
  return 0;
}


/*
 * Low-level transmit functions that actually place the packet onto
 * the channel.
 */
void
Mac802_11::sendRTS(int dst)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct rts_frame *rf = (struct rts_frame*)p->access(hdr_mac::offset_);
	
	assert(pktTx_);
	assert(pktRTS_ == 0);

	/*
	 *  If the size of the packet is larger than the
	 *  RTSThreshold, then perform the RTS/CTS exchange.
	 */
	if( (u_int32_t) HDR_CMN(pktTx_)->size() < macmib_.getRTSThreshold() ||
            (u_int32_t) dst == MAC_BROADCAST) {
		Packet::free(p);
		return;
	}

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = phymib_.getRTSlen();
	ch->iface() = -2;
	ch->error() = 0;

	bzero(rf, MAC_HDR_LEN);

	rf->rf_fc.fc_protocol_version = MAC_ProtocolVersion;
 	rf->rf_fc.fc_type	= MAC_Type_Control;
 	rf->rf_fc.fc_subtype	= MAC_Subtype_RTS;
 	rf->rf_fc.fc_to_ds	= 0;
 	rf->rf_fc.fc_from_ds	= 0;
 	rf->rf_fc.fc_more_frag	= 0;
 	rf->rf_fc.fc_retry	= 0;
 	rf->rf_fc.fc_pwr_mgt	= netif_->PowerSave()? 1 : 0;
 	rf->rf_fc.fc_more_data	= 0;
 	rf->rf_fc.fc_wep	= 0;
 	rf->rf_fc.fc_order	= 0;

	//rf->rf_duration = RTS_DURATION(pktTx_);
	STORE4BYTE(&dst, (rf->rf_ra));
	
	/* store rts tx time */
 	ch->txtime() = txtime(ch->size(), basicRate_);
	
	STORE4BYTE(&index_, (rf->rf_ta));

	/* calculate rts duration field */	
	rf->rf_duration = usec(phymib_.getSIFS()
			       + txtime(phymib_.getCTSlen(), basicRate_)
			       + phymib_.getSIFS()
                               + txtime(pktTx_)
			       + phymib_.getSIFS()
			       + txtime(phymib_.getACKlen(), basicRate_));
	pktRTS_ = p;
}

void
Mac802_11::sendCTS(int dst, double rts_duration)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct cts_frame *cf = (struct cts_frame*)p->access(hdr_mac::offset_);

	assert(pktCTRL_ == 0);

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	ch->size() = phymib_.getCTSlen();

	ch->iface() = -2;
	ch->error() = 0;
	//ch->direction() = hdr_cmn::DOWN;
	bzero(cf, MAC_HDR_LEN);

	cf->cf_fc.fc_protocol_version = MAC_ProtocolVersion;
	cf->cf_fc.fc_type	= MAC_Type_Control;
	cf->cf_fc.fc_subtype	= MAC_Subtype_CTS;
 	cf->cf_fc.fc_to_ds	= 0;
 	cf->cf_fc.fc_from_ds	= 0;
 	cf->cf_fc.fc_more_frag	= 0;
 	cf->cf_fc.fc_retry	= 0;
 	cf->cf_fc.fc_pwr_mgt	= netif_->PowerSave()? 1 : 0;
 	cf->cf_fc.fc_more_data	= 0;
 	cf->cf_fc.fc_wep	= 0;
 	cf->cf_fc.fc_order	= 0;
	
	//cf->cf_duration = CTS_DURATION(rts_duration);
	STORE4BYTE(&dst, (cf->cf_ra));
	
	/* store cts tx time */
	ch->txtime() = txtime(ch->size(), basicRate_);
	
	/* calculate cts duration */
	cf->cf_duration = usec(sec(rts_duration)
                              - phymib_.getSIFS()
                              - txtime(phymib_.getCTSlen(), basicRate_));

	pktCTRL_ = p;
}

void
Mac802_11::sendACK(int dst)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct ack_frame *af = (struct ack_frame*)p->access(hdr_mac::offset_);

	assert(pktCTRL_ == 0);

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	// CHANGE WRT Mike's code
	ch->size() = phymib_.getACKlen();
	ch->iface() = -2;
	ch->error() = 0;
	
	bzero(af, MAC_HDR_LEN);

	af->af_fc.fc_protocol_version = MAC_ProtocolVersion;
 	af->af_fc.fc_type	= MAC_Type_Control;
 	af->af_fc.fc_subtype	= MAC_Subtype_ACK;
 	af->af_fc.fc_to_ds	= 0;
 	af->af_fc.fc_from_ds	= 0;
 	af->af_fc.fc_more_frag	= 0;
 	af->af_fc.fc_retry	= 0;
 	af->af_fc.fc_pwr_mgt	= netif_->PowerSave()? 1 : 0;
 	af->af_fc.fc_more_data	= 0;
 	af->af_fc.fc_wep	= 0;
 	af->af_fc.fc_order	= 0;

	//af->af_duration = ACK_DURATION();
	STORE4BYTE(&dst, (af->af_ra));

	/* store ack tx time */
 	ch->txtime() = txtime(ch->size(), basicRate_);
	
	/* calculate ack duration */
 	af->af_duration = 0;	
	
	pktCTRL_ = p;
}

void
Mac802_11::sendDATA(Packet *p)
{
	hdr_cmn* ch = HDR_CMN(p);
	struct hdr_mac802_11* dh = HDR_MAC802_11(p);

	assert(pktTx_ == 0);

	/*
	 * Update the MAC header
	 */
	ch->size() += phymib_.getHdrLen11();

	dh->dh_fc.fc_protocol_version = MAC_ProtocolVersion;
	dh->dh_fc.fc_type       = MAC_Type_Data;
	dh->dh_fc.fc_subtype    = MAC_Subtype_Data;
	
	dh->dh_fc.fc_to_ds      = 0;
	dh->dh_fc.fc_from_ds    = 0;
	dh->dh_fc.fc_more_frag  = 0;
	dh->dh_fc.fc_retry      = 0;
	dh->dh_fc.fc_pwr_mgt    = netif_->PowerSave()? 1 : 0;
	dh->dh_fc.fc_more_data  = 0;
	dh->dh_fc.fc_wep        = 0;
	dh->dh_fc.fc_order      = 0;

	/* store data tx time */
 	ch->txtime() = txtime(ch->size(), dataRate_);

	if((u_int32_t)ETHER_ADDR(dh->dh_ra) != MAC_BROADCAST) {
		/* store data tx time for unicast packets */
		ch->txtime() = txtime(ch->size(), dataRate_);
		
		dh->dh_duration = usec(txtime(phymib_.getACKlen(), basicRate_)
				       + phymib_.getSIFS());



	} else {
		/* store data tx time for broadcast packets (see 9.6) */
		ch->txtime() = txtime(ch->size(), basicRate_);
		
		dh->dh_duration = 0;
	}
	pktTx_ = p;
}

void
Mac802_11::sendBEACON()
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct hdr_mac802_11* dh = HDR_MAC802_11(p);
	struct hdr_beacon *bh = hdr_beacon::access(p);

	/*
	 * Update the MAC header
	 */
	bzero(dh, MAC_HDR_LEN);

	dh->dh_fc.fc_protocol_version = MAC_ProtocolVersion;
	dh->dh_fc.fc_type       = MAC_Type_Management;
	dh->dh_fc.fc_subtype    = MAC_Subtype_Beacon;
	dh->dh_fc.fc_to_ds      = 0;
	dh->dh_fc.fc_from_ds    = 0;
	dh->dh_fc.fc_more_frag  = 0;
	dh->dh_fc.fc_retry      = 0;
	dh->dh_fc.fc_pwr_mgt    = netif_->PowerSave()? 1 : 0;
	dh->dh_fc.fc_more_data  = 0;
	dh->dh_fc.fc_wep        = 0;
	dh->dh_fc.fc_order      = 0;
	dh->dh_duration 	= 0;

	uint32_t dst = MAC_BROADCAST;
	STORE4BYTE(&index_, dh->dh_ta);
	STORE4BYTE(&dst, dh->dh_ra);

	 // beacon header
	bzero(bh, sizeof(struct hdr_beacon));
	// bh->bh_bb.bb_ts = tsfTimer.getTSF();
	// note: tsf is filled in later
	bh->bh_bb.bb_bi = (u_int16_t) (macmib_.BeaconPeriod * 1e3); // Kus
	// XXX settings for STAs in an IBSS only
	bh->bh_bb.bb_ci.ci_ess = (u_char) 0x0;
	bh->bh_bb.bb_ci.ci_ibss = (u_char) 0x1;
	bh->bh_bb.bb_ci.ci_cfpollable = (u_char) 0x0;
	bh->bh_bb.bb_ci.ci_cfpollreq = (u_char) 0x0;
	// XXX assume no WEP for now
	bh->bh_bb.bb_ci.ci_privacy = (u_char) 0x0;
	bh->bh_bb.bb_ssid.ssid_eid = (u_int8_t) 0x0;
	bh->bh_bb.bb_ssid.ssid_length = (u_int8_t) 0x1;
	// we set the SSID to zero so all stations in the simulation
	// are in the same universe.
	bh->bh_bb.bb_sr.sr_eid = (u_int8_t) 0x0;
	bh->bh_bb.bb_sr.sr_length = (u_int8_t) 0x0;
	bh->bh_bb.bb_sr.sr_rates[0] = (u_char) 0x12;
	// XXX bh->bh_bb.bb_fps.fps_fps_eid
	bh->bh_bb.bb_fps.fps_length = (u_int8_t) 1;
	bh->bh_bb.bb_fps.fps_dwelltime = (u_int16_t) 390;
	bh->bh_bb.bb_fps.fps_hopset = (u_int8_t) 1;
	// bh->bh_bb.bb_fps.fps_hoppat
	// bh->bh_bb.bb_fps.fps_hopind
	// XXX don't do indices for now
	bh->bh_bb.bb_dps = 0;
	// note: cf_param_set unused for PCF only hence unused
	bh->bh_bb.bb_ips.ips_eid = 0;
	bh->bh_bb.bb_ips.ips_length = (u_int8_t) 2;
	bh->bh_bb.bb_ips.ips_atim_wind = 
	       (u_int16_t) (macmib_.ATIMWindow * 1e3);
	//XXX p. 135 is unclear
	bh->bh_bb.bb_tim.tim_eid = (u_int8_t) 0x0;
	// note: TIM unused for DCF STAs
	// bh->_bh_fcs: FCS unimplemented (but takes up space) XXXXX

	// common header
	ch->ptype() = PT_MAC;
	ch->iface() = -2;
	ch->error() = 0;
	ch->size() = phymib_.getBeaconLen(bh);
	ch->txtime() = txtime(ch->size(), basicRate_);

	pktBEACON_ = p;
}



/* ======================================================================
   Retransmission Routines
   ====================================================================== */
void
Mac802_11::RetransmitRTS()
{
	assert(pktTx_);
	assert(pktRTS_);

	if (mhEndATIM_.busy()) {
		discard(pktRTS_, DROP_MAC_INVALID_STATE);
		pktRTS_ = 0;
		// buffer the packet and do nothing for now
		hdr_cmn *ch = HDR_CMN(pktTx_);
		ch->size() -= phymib_.getHdrLen11();
		struct buffer_entry *new_entry = 
		   new buffer_entry(pktTx_, false, false, false, 0);
		CIRCLEQ_INSERT_TAIL(&buf_head_, new_entry, entries);
		pktTx_ = 0;
		return;
	}

	assert(mhBackoff_.busy() == 0);
	macmib_.RTSFailureCount++;

	ssrc_ += 1;			// STA Short Retry Count
		
	if(ssrc_ >= macmib_.getShortRetryLimit()) {
		discard(pktRTS_, DROP_MAC_RETRY_COUNT_EXCEEDED); pktRTS_ = 0;
		/* tell the callback the send operation failed 
		   before discarding the packet */
		hdr_cmn *ch = HDR_CMN(pktTx_);
		if (ch->xmit_failure_) {
                        /*
                         *  Need to remove the MAC header so that 
                         *  re-cycled packets don't keep getting
                         *  bigger.
                         */
			ch->size() -= phymib_.getHdrLen11();
                        ch->xmit_reason_ = XMIT_REASON_RTS;
                        ch->xmit_failure_(pktTx_->copy(),
                                          ch->xmit_failure_data_);
                }
		discard(pktTx_, DROP_MAC_RETRY_COUNT_EXCEEDED); 
		pktTx_ = 0;
		ssrc_ = 0;
		rst_cw();
	} else {
		struct rts_frame *rf;
		rf = (struct rts_frame*)pktRTS_->access(hdr_mac::offset_);
		rf->rf_fc.fc_retry = 1;

		inc_cw();
		mhBackoff_.start(cw_, is_idle());
	}
}

void
Mac802_11::RetransmitDATA()
{
	struct hdr_cmn *ch;
	struct hdr_mac802_11 *mh;
	u_int32_t *rcount, thresh;

	assert(pktTx_);
	assert(pktRTS_ == 0);

	if (mhEndATIM_.busy()) {
		// buffer the packet and do nothing for now
		hdr_cmn *ch = HDR_CMN(pktTx_);
		ch->size() -= phymib_.getHdrLen11();
		struct buffer_entry *new_entry = 
		   new buffer_entry(pktTx_, false, false, false, 0);
		CIRCLEQ_INSERT_TAIL(&buf_head_, new_entry, entries);
		pktTx_ = 0;
		return;
	}

	assert(mhBackoff_.busy() == 0);
	ch = HDR_CMN(pktTx_);
	mh = HDR_MAC802_11(pktTx_);

	/*
	 *  Broadcast packets don't get ACKed and therefore
	 *  are never retransmitted.
	 */
	if((u_int32_t)ETHER_ADDR(mh->dh_ra) == MAC_BROADCAST) {
		Packet::free(pktTx_); 
		pktTx_ = 0;

		/*
		 * Backoff at end of TX.
		 */
		rst_cw();
		mhBackoff_.start(cw_, is_idle());

		return;
	}

	macmib_.ACKFailureCount++;

	if((u_int32_t) ch->size() <= macmib_.getRTSThreshold()) {
                rcount = &ssrc_;
               thresh = macmib_.getShortRetryLimit();
        } else {
                rcount = &slrc_;
               thresh = macmib_.getLongRetryLimit();
        }

	(*rcount)++;

	if(*rcount >= thresh) {
		/* IEEE Spec section 9.2.3.5 says this should be greater than
		   or equal */
		macmib_.FailedCount++;
		/* tell the callback the send operation failed 
		   before discarding the packet */
		hdr_cmn *ch = HDR_CMN(pktTx_);
		if (ch->xmit_failure_) {
                        ch->size() -= phymib_.getHdrLen11();
			ch->xmit_reason_ = XMIT_REASON_ACK;
                        ch->xmit_failure_(pktTx_->copy(),
                                          ch->xmit_failure_data_);
                }

		discard(pktTx_, DROP_MAC_RETRY_COUNT_EXCEEDED); 
		pktTx_ = 0;
		*rcount = 0;
		rst_cw();
	}
	else {
		struct hdr_mac802_11 *dh;
		dh = HDR_MAC802_11(pktTx_);
		dh->dh_fc.fc_retry = 1;

		sendRTS(ETHER_ADDR(mh->dh_ra));
		inc_cw();
		mhBackoff_.start(cw_, is_idle());
	}
}


void Mac802_11::RetransmitATIM()

{
  struct hdr_mac802_11 *mh = HDR_MAC802_11(pktATIM_);;

  assert(mhBackoff_.busy() == 0);
  assert(pktATIM_);

  if ((!mhEndATIM_.busy()) || (ETHER_ADDR(mh->dh_ra) == MAC_BROADCAST)) {
    Packet::free(pktATIM_);
    pktATIM_ = 0;
    rst_cw();
    mhBackoff_.start(cw_, is_idle());
    // mhBackoff_.start(cw, 0.0, is_idle(), "rexmit atim backoff");
    return;
  }

  // look to see if this is still necessary 
  struct buffer_entry *this_entry;
  CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
    if ((this_entry->addr(this) == ETHER_ADDR(mh->dh_ra)) &&
	(this_entry->recvd_ack_ || !this_entry->pwr_mgt_)) {
      Packet::free(pktATIM_);
      pktATIM_ = 0;
      rst_cw();
      mhBackoff_.start(cw_, is_idle());
      // mhBackoff_.start(cw, 0.0, is_idle(), "rexmit atim backoff");
      return;
    }
  }

  macmib_.ACKFailureCount++;
  atimrc_++;
  if (atimrc_ > MAC_ATIMRetryLimit) {
    macmib_.FailedCount++;
    discard(pktATIM_, DROP_MAC_ATIM_RETRY_COUNT);
    pktATIM_ = 0;
    atimrc_ = 0;
    rst_cw();
    
    CIRCLEQ_HEAD(buffer_head, buffer_entry) expired;
    CIRCLEQ_INIT(&expired);

    CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
      if (this_entry->addr(this) == ETHER_ADDR(last_atim_ra_)) { 
	CIRCLEQ_REMOVE(&buf_head_, this_entry, entries); 
        CIRCLEQ_INSERT_TAIL(&expired, this_entry, entries);
      }
    }

    CIRCLEQ_FOREACH(this_entry, &expired, entries) { 
      if (!this_entry->sent_data_) { 
        hdr_cmn *ch = HDR_CMN(this_entry->packet_); 
        if (ch->xmit_failure_) { 
	  ch->xmit_reason_ = XMIT_REASON_ATIM;
	  ch->xmit_failure_
	    (this_entry->packet_->copy(), ch->xmit_failure_data_); 
        } 
        discard(this_entry->packet_, DROP_MAC_RETRY_COUNT_EXCEEDED);
      }
      if (this_entry == entryATIM_) entryATIM_ = 0;
      delete this_entry;
    }
  } else {
    // retry count is at most threshold
    mh->dh_fc.fc_retry = 1;
    inc_cw();
    mhBackoff_.start(cw_, is_idle());
    // mhBackoff.start(cw, 0.0, is_idle(), "rexmit data backoff - 2");
  }

}


/* ======================================================================
   Incoming Packet Routines
   ====================================================================== */
void
Mac802_11::send(Packet *p, Handler *h)
{
	double rTime;
	struct hdr_mac802_11* dh = HDR_MAC802_11(p);
	struct hdr_span *spanh = hdr_span::access(p);
	struct buffer_entry *this_entry = 0;
	struct hdr_cmn* ch = HDR_CMN(p);
	bool fastroute = spanh->to_coordinator;
	int entry = 0;

	if (netif_->state() == pwr_state_off) {
	  Packet::free(p);
	  return;
	}

	EnergyModel *em = netif_->node()->energy_model();
	if (em && em->sleep()) {
		em->set_node_sleep(0);
		em->set_node_state(EnergyModel::INROUTE);
	}

	if (idle_start > 0.01) {
	   	total_true_idle_time += 
		  (Scheduler::instance().clock() - idle_start);
		idle_start = 0.0;
	}
	
	callback_ = h;

	/* Assign the data packet a sequence number */
	dh->dh_scontrol = sta_seqno_++;

	/* correct to_coordinator field in Span header if we
	 * have more up to date pwr mgt status */
	CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) { 
	  if (this_entry->addr(this) == ETHER_ADDR(dh->dh_ra)) {
	    if (this_entry->pwr_mgt_updated_) {
	       fastroute = !this_entry->pwr_mgt_;
	       break;
	    }
	  }
	}

	if (!psm_mode_) {
	  assert(pktTx_ == 0);
	  assert((netif_->state() == pwr_state_idle) ||
	   (netif_->state() == pwr_state_off));
	  assert(!(netif_->PowerSave()));
	}
  
	if (netif_->state() != pwr_state_idle) {
	  struct buffer_entry *new_entry =
	     new buffer_entry(p, false, false, false, 0);
	  new_entry->pwr_mgt_ = !fastroute;
	  CIRCLEQ_INSERT_TAIL(&buf_head_, new_entry, entries);
	  return;
	}

	if (mhEndATIM_.busy() && got_beacon_) { // inside ATIM window
	  struct buffer_entry *new_entry =
	    new buffer_entry(p, false, false, false, 0);
	  bool found = false;

	  // check all ATIM entries to same next hop
	  CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
	    if ((this_entry->addr(this) == new_entry->addr(this)) &&
	      (this_entry->sent_atim_ || !this_entry->pwr_mgt_)) {
	      new_entry->sent_atim_ = this_entry->sent_atim_;
	      new_entry->recvd_ack_ = this_entry->recvd_ack_;
              new_entry->pwr_mgt_ = !fastroute;
	      found = true;
	      break;
	    }
	  }
	  CIRCLEQ_INSERT_TAIL(&buf_head_, new_entry, entries);

	  // send an atim only if not gratuitous
	  if (!found && !pktATIM_) {
	    atim_scan();

	    if (pktATIM_) {
	      if (first_atim_) {
	        if (!mhBackoff_.busy() && !mhDefer_.busy() && 
	            !mhSend_.busy())
	              mhBackoff_.start(cw_, 1);
	          //  mhBackoff_.start(cw, 0.0, 1, "first atim backoff");
	        first_atim_ = false;

	      } else {  // !first_atim_
	        if (!mhSend_.busy() && !mhBackoff_.busy()) {
	          if (is_idle()) {
	            if (mhDefer_.busy() == 0) {
	               rTime = (Random::random() % cw_)
	                   * (phymib_.getSlotTime());
	               mhDefer_.start(phymib_.getDIFS() + rTime);
	            }
	          } else {
	            mhBackoff_.start(cw_, is_idle());
	        //  mhBackoff.start(cw, 0.0, is_idle(), "backoff (send)");
	          }
	        }
	      }
	    }
	  } else callback();
	  // done with ATIM window
	  return;
	}

	// we are outside the ATIM window, or inside the atim window 
	// and don't have a beacon yet: search for the station in acked 
	// or bdcast pkts.
#if SPAN_DEBUG > 1
	fprintf(stderr,
	    "%f: %d sending outside atim window or not got beacon.\n",
	    Scheduler::instance().clock(), index);
#endif

	bool found = false;
	CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
	  if ((this_entry->addr(this) == ETHER_ADDR(dh->dh_ra)) &&
	      (this_entry->recvd_ack_ ||
	       (ETHER_ADDR(dh->dh_ra) == MAC_BROADCAST &&
	        this_entry->sent_atim_) ||
	       (!this_entry->pwr_mgt_ && fastroute))) { 
	    found = true;
	    break;
	  }
	}
    
	if ((!found && !fastroute && psm_mode_) || pktTx_ || 
	     mhEndATIM_.busy()) {
      // no atim was found, or we are inside the atim window waiting for a
      // beacon, or there is a packet pending transmission, or if fast route
      // is not on...
#if SPAN_DEBUG > 1
	  fprintf(stderr, "%f: %d buffering the pkt (found %d, fr %d): ",
	      Scheduler::instance().clock(), index, found, fastroute);
	  trace_pkt(p);
#endif
	  // buffer the new packet and do nothing for now
	  struct buffer_entry *new_entry =
	    new buffer_entry(p, false, false, false, 0);
	  new_entry->pwr_mgt_ = !fastroute;
	  CIRCLEQ_INSERT_TAIL(&buf_head_, new_entry, entries);

	} else {
	// an atim was found and we are outside the atim window; 
	// send data with the dcf, OR we are doing fast route
#if SPAN_DEBUG > 1
	  fprintf(stderr, "%f: %d sending with the dcf\n",
	      Scheduler::instance().clock(), index);
#endif
	  sendDATA(p);
	  sendRTS(ETHER_ADDR(dh->dh_ra));

	  if (fastroute && !found) {
#if SPAN_DEBUG > 2
	    fprintf(stderr, "%f: %d optimistic send to coordinator %d\n",
	      Scheduler::instance().clock(),
	      index, ETHER_ADDR(dh->dh_ra));
#endif
	  }

	/* If the medium is IDLE, we must wait for a 
	 * DIFS Space before transmitting.
	 */
	  if (!mhSend_.busy() && !mhBackoff_.busy()) {
	    if (is_idle()) {
	    /* If we are already deferring, there is no
	     * need to reset the Defer timer. */
	      if (mhDefer_.busy() == 0) {
	        rTime = (Random::random() % cw_)
	                 * (phymib_.getSlotTime());
	        mhDefer_.start(phymib_.getDIFS() + rTime);
	      }
	    }

	    /* If the medium is NOT IDLE, then we start
	     * the backoff timer. */
	    else {
	      // mhBackoff.start(cw, 0.0, is_idle(), "backoff (send)");
	      mhBackoff_.start(cw_, is_idle());
	    }
	  }
	}
}

void
Mac802_11::recv(Packet *p, Handler *h)
{
	struct hdr_cmn *hdr = HDR_CMN(p);
	/*
	 * Sanity Check
	 */
	assert(initialized());

	if (hdr->uid() == 805) {
		printf("Just hit packet 805!\n");
	}

	/*
	 *  Handle outgoing packets.
	 */
        // If a packet gets here and doesn't have a direction yet, 
        // warn user
	if(hdr->direction() == hdr_cmn::NONE) {
		fprintf(stderr, "Packet without direction, check sender!\n");
	}

	if(hdr->direction() == hdr_cmn::DOWN) {
                send(p, h);
                return;
        }

	if (netif_->state() == pwr_state_sleep) { 
		printf("Receiving packet while asleep!\n");
	}
	assert(netif_->state() == pwr_state_idle);

	/*
	 *  Handle incoming packets.
	 *
	 *  We just received the 1st bit of a packet on the network
	 *  interface.
	 *
	 */

	/*
	 *  If the interface is currently in transmit mode, then
	 *  it probably won't even see this packet.  However, the
	 *  "air" around me is BUSY so I need to let the packet
	 *  proceed.  Just set the error flag in the common header
	 *  to that the packet gets thrown away.
	 */
	if(tx_active_ && hdr->error() == 0) {
		hdr->error() = 1;
	}

	if(rx_state_ == MAC_IDLE) {
		setRxState(MAC_RECV);
		pktRx_ = p;
		/*
		 * Schedule the reception of this packet, in
		 * txtime seconds.
		 */
		mhRecv_.start(txtime(p));
	} else {
		/*
		 *  If the power of the incoming packet is smaller than the
		 *  power of the packet currently being received by at least
                 *  the capture threshold, then we ignore the new packet.
		 */
		if(pktRx_->txinfo_.RxPr / p->txinfo_.RxPr >= p->txinfo_.CPThresh) {
			capture(p);
		} else {
			collision(p);
		}
	}
}

void
Mac802_11::recv_timer()
{
	u_int32_t src; 
	hdr_cmn *ch = HDR_CMN(pktRx_);
	hdr_mac802_11 *mh = HDR_MAC802_11(pktRx_);
	u_int32_t dst = ETHER_ADDR(mh->dh_ra);
	
	u_int8_t  type = mh->dh_fc.fc_type;
	u_int8_t  subtype = mh->dh_fc.fc_subtype;

	assert(pktRx_);
	assert(rx_state_ == MAC_RECV || rx_state_ == MAC_COLL);
	
        /*
         *  If the interface is in TRANSMIT mode when this packet
         *  "arrives", then I would never have seen it and should
         *  do a silent discard without adjusting the NAV.
         */
        if(tx_active_) {
                Packet::free(pktRx_);
                goto done;
        }

	/*
	 * Handle collisions.
	 */
	if(rx_state_ == MAC_COLL) {
		discard(pktRx_, DROP_MAC_COLLISION);		
		set_nav(usec(phymib_.getEIFS()));
		goto done;
	}

	/*
	 * Check to see if this packet was received with enough
	 * bit errors that the current level of FEC still could not
	 * fix all of the problems - ie; after FEC, the checksum still
	 * failed.
	 */
	if( ch->error() ) {
		Packet::free(pktRx_);
		set_nav(usec(phymib_.getEIFS()));
		goto done;
	}

	/*
	 * IEEE 802.11 specs, section 9.2.5.6
	 *	- update the NAV (Network Allocation Vector)
	 */
	if(dst != (u_int32_t)index_) {
		set_nav(mh->dh_duration);
	}

        /* tap out - */
        if (tap_ && type == MAC_Type_Data &&
            MAC_Subtype_Data == subtype ) 
		tap_->tap(pktRx_);

	/* update our power management table */
	buffer_entry *this_entry;
	switch (type) {
	case MAC_Type_Management:
	  switch (subtype) {
	  case MAC_Subtype_Beacon:
	  case MAC_Subtype_ATIM:
	    CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
	      if (this_entry->addr(this) == ETHER_ADDR(mh->dh_ta)) {
	        this_entry->pwr_mgt_ = (mh->dh_fc.fc_pwr_mgt == 1);
	        this_entry->pwr_mgt_updated_ = true;
	      }
	    }
	    break;
	  default:
	    break;
	  }
	  break;

	case MAC_Type_Control:
	  switch (subtype) {
	  case MAC_Subtype_RTS: {
	    struct rts_frame *rf = 
	      (struct rts_frame*) pktRx_->access(hdr_mac::offset_);
	    CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
	      if (this_entry->addr(this) == ETHER_ADDR(rf->rf_ta)) {
	        this_entry->pwr_mgt_ = (rf->rf_fc.fc_pwr_mgt == 1);
	        this_entry->pwr_mgt_updated_ = true;
	      }
	    }
          }
	  break;
	  case MAC_Subtype_CTS:
	    if (pktRTS_ && dst == index_) {
	      struct cts_frame *cf =
	        (struct cts_frame *) pktRx_->access(hdr_mac::offset_);
	      struct rts_frame *rf =
	        (struct rts_frame *) pktRTS_->access(hdr_mac::offset_);
	      CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
	        // check against dst addr of RTS packet
	        if (this_entry->addr(this) == ETHER_ADDR(rf->rf_ra)) {
	          this_entry->pwr_mgt_ = (cf->cf_fc.fc_pwr_mgt == 1);
	          this_entry->pwr_mgt_updated_ = true;
	        }
	      }
	    }
	    break;
	  default:
	    break;
	  }

	case MAC_Type_Data:
	  switch (subtype) {
	  case MAC_Subtype_Data:
	    CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
	      if (this_entry->addr(this) == ETHER_ADDR(mh->dh_ta)) {
	        this_entry->pwr_mgt_ = (mh->dh_fc.fc_pwr_mgt == 1);
	        this_entry->pwr_mgt_updated_ = true;
	      }
	    }
	    break;
	  default:
	    break;
	  }
	}

	/*
	 * Adaptive Fidelity Algorithm Support - neighborhood infomation 
	 * collection
	 *
	 * Hacking: Before filter the packet, log the neighbor node
	 * I can hear the packet, the src is my neighbor
	 */
	if (netif_->node()->energy_model() && 
	    netif_->node()->energy_model()->adaptivefidelity()) {
		src = ETHER_ADDR(mh->dh_ta);
		netif_->node()->energy_model()->add_neighbor(src);
	}

	/*
	 * Address Filtering
	 */
	if((dst != (u_int32_t)index_) && (dst != MAC_BROADCAST)) {
		/*
		 *  We don't want to log this event, so we just free
		 *  the packet instead of calling the drop routine.
		 */
		discard(pktRx_, "---");
		goto done;
	}

	switch(type) {

	case MAC_Type_Management:
		switch (subtype) {
		case MAC_Subtype_Beacon:
		  recvBEACON(pktRx_);
		  break;

		case MAC_Subtype_ATIM:
		  recvATIM(pktRx_);
		  break;

		default:
		  discard(pktRx_, DROP_MAC_PACKET_ERROR);
		  goto done;
		}
		break;

	case MAC_Type_Control:
		switch(subtype) {
		case MAC_Subtype_RTS:
			recvRTS(pktRx_);
			break;
		case MAC_Subtype_CTS:
			recvCTS(pktRx_);
			break;
		case MAC_Subtype_ACK:
			recvACK(pktRx_);
			break;
		default:
			fprintf(stderr,"recvTimer1:Invalid MAC Control Subtype %x\n",
				subtype);
			exit(1);
		}
		break;

	case MAC_Type_Data:
		switch(subtype) {
		case MAC_Subtype_Data:
			recvDATA(pktRx_);
			break;
		default:
			fprintf(stderr, "recv_timer2:Invalid MAC Data Subtype %x\n",
				subtype);
			exit(1);
		}
		break;
	default:
		fprintf(stderr, "recv_timer3:Invalid MAC Type %x\n", subtype);
		exit(1);
	}
 done:
	pktRx_ = 0;
	rx_resume();
}


void
Mac802_11::recvRTS(Packet *p)
{
	struct rts_frame *rf = (struct rts_frame*)p->access(hdr_mac::offset_);

	if((tx_state_ != MAC_IDLE) || mhEndATIM_.busy() ) {
		discard(p, DROP_MAC_BUSY);
		return;
	}

	/*
	 *  If I'm responding to someone else, discard this RTS.
	 */
	if(pktCTRL_) {
		discard(p, DROP_MAC_BUSY);
		return;
	}

	sendCTS(ETHER_ADDR(rf->rf_ta), rf->rf_duration);

	/*
	 *  Stop deferring - will be reset in tx_resume().
	 */
	if(mhDefer_.busy()) mhDefer_.stop();

	tx_resume();

	mac_log(p);
}

/*
 * txtime()	- pluck the precomputed tx time from the packet header
 */
double
Mac802_11::txtime(Packet *p)
{
	struct hdr_cmn *ch = HDR_CMN(p);
	double t = ch->txtime();
	if (t < 0.0) {
		drop(p, "XXX");
 		exit(1);
	}
	return t;
}

 
/*
 * txtime()	- calculate tx time for packet of size "psz" bytes 
 *		  at rate "drt" bps
 */
double
Mac802_11::txtime(double psz, double drt)
{
	double dsz = psz - phymib_.getPLCPhdrLen();
        int plcp_hdr = phymib_.getPLCPhdrLen() << 3;	
	int datalen = (int)dsz << 3;
	double t = (((double)plcp_hdr)/phymib_.getPLCPDataRate())
                                       + (((double)datalen)/drt);
	return(t);
}



void
Mac802_11::recvCTS(Packet *p)
{
	if(tx_state_ != MAC_RTS) {
		discard(p, DROP_MAC_INVALID_STATE);
		if (pktTx_) {
		  hdr_cmn *ch = HDR_CMN(pktTx_);
		  ch->size() -= phymib_.getHdrLen11();
		  struct buffer_entry *new_entry =
		    new buffer_entry(pktTx_, false, false, false, 0);
		  CIRCLEQ_INSERT_TAIL(&buf_head_, new_entry, entries);
		  pktTx_ = 0;
		}
		return;
	}

	assert(pktRTS_);
	Packet::free(pktRTS_);
	pktRTS_ = 0;

	assert(pktTx_);	
	mhSend_.stop();

	/*
	 * The successful reception of this CTS packet implies
	 * that our RTS was successful. 
	 * According to the IEEE spec 9.2.5.3, you must 
	 * reset the ssrc_, but not the congestion window.
	 */
	ssrc_ = 0;
	rtsrc_ = 0;
	rst_cw();

	tx_resume();

	mac_log(p);
}

void
Mac802_11::recvDATA(Packet *p)
{
	struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
	u_int32_t dst, src, size;
	struct hdr_cmn *ch = HDR_CMN(p);

	dst = ETHER_ADDR(dh->dh_ra);
	src = ETHER_ADDR(dh->dh_ta);
	size = ch->size();
	/*
	 * Adjust the MAC packet size - ie; strip
	 * off the mac header
	 */
	ch->size() -= phymib_.getHdrLen11();
	ch->num_forwards() += 1;

	/*
	 *  If we sent a CTS, clean up...
	 */
	if(dst != MAC_BROADCAST) {
		if(size >= macmib_.getRTSThreshold()) {
			if (tx_state_ == MAC_CTS) {
				assert(pktCTRL_);
				Packet::free(pktCTRL_); pktCTRL_ = 0;
				mhSend_.stop();
				/*
				 * Our CTS got through.
				 */
				ssrc_ = 0;
				rst_cw();
			} else {
				discard(p, DROP_MAC_BUSY);
				return;
			}
			sendACK(src);
			tx_resume();
		} else {
			/*
			 *  We did not send a CTS and there's no
			 *  room to buffer an ACK.
			 */
			if(pktCTRL_) {
				discard(p, DROP_MAC_BUSY);
				return;
			}
			sendACK(src);
			if(mhSend_.busy() == 0)
				tx_resume();
		}
	}
	
	/* ============================================================
	   Make/update an entry in our sequence number cache.
	   ============================================================ */

	/* Changed by Debojyoti Dutta. This upper loop of if{}else was 
	   suggested by Joerg Diederich <dieder@ibr.cs.tu-bs.de>. 
	   Changed on 19th Oct'2000 */

        if(dst != MAC_BROADCAST) {
                if (src < (u_int32_t) cache_node_count_) {
                        Host *h = &cache_[src];

                        if(h->seqno && h->seqno == dh->dh_scontrol) {
                                discard(p, DROP_MAC_DUPLICATE);
                                return;
                        }
                        h->seqno = dh->dh_scontrol;
                } else {
			static int count = 0;
			if (++count <= 10) {
				printf ("MAC_802_11: accessing MAC cache_ array out of range (src %u, dst %u, size %d)!\n", src, dst, cache_node_count_);
				if (count == 10)
					printf ("[suppressing additional MAC cache_ warnings]\n");
			};
		};
	}

	/*
	 *  Pass the packet up to the link-layer.
	 *  XXX - we could schedule an event to account
	 *  for this processing delay.
	 */
	
	/* in BSS mode, if a station receives a packet via
	 * the AP, and higher layers are interested in looking
	 * at the src address, we might need to put it at
	 * the right place - lest the higher layers end up
	 * believing the AP address to be the src addr! a quick
	 * grep didn't turn up any higher layers interested in
	 * the src addr though!
	 * anyway, here if I'm the AP and the destination
	 * address (in dh_3a) isn't me, then we have to fwd
	 * the packet; we pick the real destination and set
	 * set it up for the LL; we save the real src into
	 * the dh_3a field for the 'interested in the info'
	 * receiver; we finally push the packet towards the
	 * LL to be added back to my queue - accomplish this
	 * by reversing the direction!*/

	if ((bss_id() == addr()) && ((u_int32_t)ETHER_ADDR(dh->dh_ra)!= MAC_BROADCAST)&& ((u_int32_t)ETHER_ADDR(dh->dh_3a) != addr())) {
		struct hdr_cmn *ch = HDR_CMN(p);
		u_int32_t dst = ETHER_ADDR(dh->dh_3a);
		u_int32_t src = ETHER_ADDR(dh->dh_ta);
		/* if it is a broadcast pkt then send a copy up
		 * my stack also
		 */
		if (dst == MAC_BROADCAST) {
			uptarget_->recv(p->copy(), (Handler*) 0);
		}

		ch->next_hop() = dst;
		STORE4BYTE(&src, (dh->dh_3a));
		ch->addr_type() = NS_AF_ILINK;
		ch->direction() = hdr_cmn::DOWN;
	}

	uptarget_->recv(p, (Handler*) 0);

	if (span_opt_) {
	  if (dst == MAC_BROADCAST && recvd_bcast_atim_ > 0) {
	    recvd_bcast_atim_--;
	    if ((!mhAdvEnd_.busy() ||
	     (!recvd_bcast_atim_ && !recvd_ucast_atim_ && 
	      !has_packet_tx_)) && !initial_wake_count_ && 
	      netif_->PowerSave()) {
	      // no data to send nor receive, so can go to sleep
#ifdef SPAN_DEBUG
	      fprintf(stderr, "%f: _%d_ end of bcasts or adv, turn off radio?\n",
	        Scheduler::instance().clock(), index);
#endif
	      goto_sleep();
	    }
	  }
	}

}


void
Mac802_11::recvACK(Packet *p)
{	
	struct hdr_cmn *ch = HDR_CMN(p);
	buffer_entry *this_entry;
	struct ack_frame *af = 
	  (struct ack_frame *) p->access(hdr_mac::offset_);

	if((tx_state_ != MAC_SEND) && (tx_state_ != MAC_ATIM)) {
	   // this could happen if we canceled a send when a 
	   // beacon period started...
		discard(p, DROP_MAC_INVALID_STATE);
		return;
	}

	if (tx_state_ == MAC_ATIM) {
	  atimrc_ = 0;
	  // mark buffered packets to the sender as acked
	  CIRCLEQ_FOREACH(this_entry, &buf_head_, entries) {
	    if (this_entry->addr(this) == ETHER_ADDR(last_atim_ra_)) {
	      this_entry->recvd_ack_ = true;
	      this_entry->pwr_mgt_ = (af->af_fc.fc_pwr_mgt == 1);
	      this_entry->pwr_mgt_updated_ = true;
	      has_packet_tx_++;
#if SPAN_DEBUG > 1
	      fprintf(stderr, "%f: %d got ack for ATIM %d, from %d\n",
		Scheduler::instance().clock(), index,
		has_packet_tx_, ETHER_ADDR(last_atim_ra_));
#endif
	    }
	  }
	}

	/*
	 * Inform the upper layer that we have successfully sent
	 * a packet
	 */
	if (pktTx_ && tx_state_ != MAC_ATIM) {
	  hdr_cmn *Txch = HDR_CMN(pktTx_);
	  if (Txch->xmit_success_) {
	    /*  Need to remove the MAC header so that re-cycled packets don't
	     *  keep getting bigger.  */
	    Txch->size() -= phymib_.getHdrLen11();
	    Txch->xmit_reason_ = 0;
	    Txch->xmit_success_(pktTx_->copy(), Txch->xmit_success_data_);
	  }
	}

	if (tx_state_ == MAC_SEND) {
	  Packet::free(pktTx_);
	  pktTx_ = 0;
	} else {
	  assert(tx_state_ == MAC_ATIM);
	  Packet::free(pktATIM_);
	  pktATIM_ = 0;
	}

	// assert(pktTx_);
	mhSend_.stop();

	/*
	 * The successful reception of this ACK packet implies
	 * that our DATA transmission was successful.  Hence,
	 * we can reset the Short/Long Retry Count and the CW.
	 *
	 * need to check the size of the packet we sent that's being
	 * ACK'd, not the size of the ACK packet.
	 */
	if((u_int32_t) ch->size() <= macmib_.getRTSThreshold())
		ssrc_ = 0;
	else
		slrc_ = 0;
	
	/*
	 * Backoff before sending again.
	 */
	rst_cw();
	if (mhBackoff_.busy())  mhBackoff_.stop();
	mhBackoff_.start(cw_, is_idle());

	tx_resume();

	mac_log(p);
}

void Mac802_11::recvBEACON(Packet * p)
{
  u_int64_t tsf;

#ifdef DEBUG
  fprintf(stderr, "%f: %d got a beacon\n",
	  Scheduler::instance().clock(), index);
#endif
  
  got_beacon_ = true;

  if (mhBeacon_.busy())
    mhBeacon_.stop();

  if (pktBEACON_) {
    Packet::free(pktBEACON_);
    pktBEACON_ = 0;
  }
  
  // extract headers
  struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
  struct hdr_beacon *bh = hdr_beacon::access(p);

  // verify that we really have a beacon
  assert((dh->dh_fc.fc_type == MAC_Type_Management) &&
	 (dh->dh_fc.fc_subtype == MAC_Subtype_Beacon));

  // adjust the received timestamp value by adding the receiving STA's delay
  // through its local PHY components plus time since first bit of timestamp
  // arrived at the MAC/PHY interface.
  
  // if this beacon's timestamp is later than our own (plus hysteresis),
  // adopt its timestamp.

  tsf = tsfTimer_.getTSF();
  if (bh->bh_bb.bb_ts > tsf + 10000) {
    // adopt the beacon
    tsfTimer_.setTSF(bh->bh_bb.bb_ts);
    tsf = tsfTimer_.getTSF();

    if (mhTBTT_.busy())
      mhTBTT_.stop();
  
    // backoff to send the first atim
    if (!mhSend_.busy() && !mhBackoff_.busy())
      mhBackoff_.start(cw_, is_idle());
      // mhBackoff_.start(cw, 0.0, is_idle(), "first atim backoff");
    got_beacon_ = true;
    double tbtt = macmib_.BeaconPeriod -
	fmod((double) tsf, macmib_.BeaconPeriod * 1e6) / 1e6;
    if (tbtt < macmib_.BeaconPeriod / 3) {
      // mhTBTT_.start(tbtt + macmib_.BeaconPeriod, "tbtt recvBEACON");
      mhTBTT_.start(tbtt + macmib_.BeaconPeriod);
    } else {
      // mhTBTT_.start(tbtt, "tbtt recvBEACON");
      mhTBTT_.start(tbtt);
    }
  }
  
  if (!mhSend_.busy()) /* there may be other DATA or CTRL pkts yet
			 to be acknowledged, so only try to transmit
			 if mhSend is not busy */
    tx_resume();

  mac_log(p);
}


void Mac802_11::recvATIM(Packet * p)
{
  recvd_atim_ = true;

  struct hdr_mac802_11 *dh = HDR_MAC802_11(p);
  u_int32_t sendaddr = ETHER_ADDR(dh->dh_ta);
  u_int32_t destaddr = ETHER_ADDR(dh->dh_ra);

  if (destaddr == MAC_BROADCAST) { // broadcast ATIM
    recvd_bcast_atim_++;
  } else { // unicast ATIM
    recvd_ucast_atim_++;
#if SPAN_DEBUG > 1
    fprintf(stderr, "%f: %d received an ATIM and is acking %d %d\n",
	    Scheduler::instance().clock(), index,
	    sendaddr, recvd_ucast_atim_);
#endif
    if (!pktCTRL_) {
      sendACK(sendaddr);
      if (mhSend_.busy() == 0)
	tx_resume();
    } 
  }

  mac_log(p);
}

int Mac802_11::i_am_idle()
{
  if ((tx_active_) || (rx_state_ != MAC_IDLE)) {
    return 0;
  }
  if (tx_state_ != MAC_IDLE)
    return 0;

  if (pktCTRL_)
    return 0;
 
  if (!is_idle())
    return 0;

  return 1;
}


void Mac802_11::handleBackoffInterrupt()
{
  if (!mhBackoff_.busy()) { return; }

  if (index_ == (u_int32_t) (God::instance()->nodes() - 1))
    return;

  Mac802_11 *nextnode =
      (Mac802_11 *) (God::instance()->getMacList()[index_ + 1]);
  if (is_idle() && (nextnode->i_am_idle())) {
    total_wasted_time += phymib_.getSlotTime();
  }
}

