/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*-
 *
 * Copyright (c) 1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Header: /nfs/jade/vint/CVSROOT/ns-2/mac/mac-802_11.h,v 1.24 2004/04/02 01:00:25 xuanc Exp $
 *
 * Ported from CMU/Monarch's code, nov'98 -Padma.
 * wireless-mac-802_11.h
 */

#ifndef ns_mac_80211_h
#define ns_mac_80211_h

// Added by Sushmita to support event tracing (singal@nunki.usc.edu)
#include "address.h"
#include "ip.h"

#include "mac-timers.h"
#include "mac/tsftimer.h"
#include "marshall.h"
#include <math.h>
#include "queue/queue.h"
#include "mac/mac.h"
// #include "mac-802_11-mlme.h"

class EventTrace;

#define DROP_MAC_BUFFER_FULL "BUF"

#define GET_ETHER_TYPE(x)	GET2BYTE((x))
#define SET_ETHER_TYPE(x,y)	{u_int16_t t = (y); STORE2BYTE(x,&t);}
#define PWRMGR()		((Wireless_Phy *)netif_)

/* ======================================================================
   Frame Formats
   ====================================================================== */

#define	MAC_ProtocolVersion	0x00

#define MAC_Type_Management	0x00
#define MAC_Type_Control	0x01
#define MAC_Type_Data		0x02
#define MAC_Type_Reserved	0x03

#define MAC_Subtype_ATIM        0x09
#define MAC_Subtype_Beacon      0x0A
#define MAC_Subtype_RTS		0x0B
#define MAC_Subtype_CTS		0x0C
#define MAC_Subtype_ACK		0x0D
#define MAC_Subtype_Data	0x00


struct frame_control {
	u_char		fc_subtype		: 4;
	u_char		fc_type			: 2;
	u_char		fc_protocol_version	: 2;

	u_char		fc_order		: 1;
	u_char		fc_wep			: 1;
	u_char		fc_more_data		: 1;
	u_char		fc_pwr_mgt		: 1;
	u_char		fc_retry		: 1;
	u_char		fc_more_frag		: 1;
	u_char		fc_from_ds		: 1;
	u_char		fc_to_ds		: 1;
};

struct rts_frame {
	struct frame_control	rf_fc;
	u_int16_t		rf_duration;
	u_char			rf_ra[ETHER_ADDR_LEN];
	u_char			rf_ta[ETHER_ADDR_LEN];
	u_char			rf_fcs[ETHER_FCS_LEN];
};

struct cts_frame {
	struct frame_control	cf_fc;
	u_int16_t		cf_duration;
	u_char			cf_ra[ETHER_ADDR_LEN];
	u_char			cf_fcs[ETHER_FCS_LEN];
};

struct ack_frame {
	struct frame_control	af_fc;
	u_int16_t		af_duration;
	u_char			af_ra[ETHER_ADDR_LEN];
	u_char			af_fcs[ETHER_FCS_LEN];
};

// XXX This header does not have its header access function because it shares
// the same header space with hdr_mac.
struct hdr_mac802_11 {
	struct frame_control	dh_fc;
	u_int16_t		dh_duration;
	u_char                  dh_ra[ETHER_ADDR_LEN]; // receiver addr
        u_char                  dh_ta[ETHER_ADDR_LEN]; // transmitter addr
        u_char                  dh_3a[ETHER_ADDR_LEN]; // bssid
	u_int16_t		dh_scontrol;
	u_char			dh_body[0]; // XXX Non-ANSI
};


/* ======================================================================
   Definitions
   ====================================================================== */

/*
 *  IEEE 802.11 Spec, section 9.2.5.7
 *	- After transmitting an RTS, a node waits CTSTimeout
 *	  seconds for a CTS.
 *
 *  IEEE 802.11 Spec, section 9.2.8
 *	- After transmitting DATA, a node waits ACKTimeout
 *	  seconds for an ACK.
 *
 *  IEEE 802.11 Spec, section 9.2.5.4
 *	- After hearing an RTS, a node waits NAVTimeout seconds
 *	  before resetting its NAV.  I've coined the variable
 *	  NAVTimeout.
 *
 */

/*  These shouldn't be needed
#define CTSTimeout	((RTS_Time + CTS_Time) + 2 * sifs)
#define ACKTimeout(len)	(DATA_Time(len) + ACK_Time + sifs + difs)
#define NAVTimeout	(2 * phymib_->SIFSTime + CTS_Time + \
                         2 * phymib->SlotTime)
#define RTS_DURATION(pkt)	\
	usec(sifs + CTS_Time + sifs + txtime(pkt) + sifs + ACK_Time)
#define CTS_DURATION(d)		\
	usec((d * 1e-6) - (CTS_Time + sifs))
#define DATA_DURATION()		\
	usec(ACK_Time + sifs)
#define ACK_DURATION()	0x00		// we're not doing fragments now
*/


/*
 * IEEE 802.11 Spec, section 15.3.2
 *	- default values for the DSSS PHY MIB
 */
#define DSSS_CWMin			31
#define DSSS_CWMax			1023
#define DSSS_SlotTime			0.000020	// 20us
#define	DSSS_CCATime			0.000015	// 15us
#define DSSS_RxTxTurnaroundTime		0.000005	// 5us
#define DSSS_SIFSTime			0.000010	// 10us
#define DSSS_PreambleLength		144		// 144 bits
#define DSSS_PLCPHeaderLength		48		// 48 bits

/* Must account for propagation delays added by the channel model when
 * calculating tx timeouts (as set in tcl/lan/ns-mac.tcl).
 *   -- Gavin Holland, March 2002
 */
#define DSSS_MaxPropagationDelay        0.000002        // 2us   XXXX

class PHY_MIB {
public:
	PHY_MIB(Mac802_11 *parent);

	inline u_int32_t getCWMin() { return(CWMin); }
        inline u_int32_t getCWMax() { return(CWMax); }
	inline double getSlotTime() { return(SlotTime); }
	inline double getCCATime() { return(CCATime); }
	inline double getRxTxTurnaroundTime() { return (RxTxTurnaroundTime); }
	inline double getSIFS() { return(SIFSTime); }
	inline double getPIFS() { return(SIFSTime + SlotTime); }
	inline double getDIFS() { return(SIFSTime + 2 * SlotTime); }
	inline double getEIFS() {
		// see (802.11-1999, 9.2.10)
		return(SIFSTime + getDIFS()
                       + (8 *  getACKlen())/PLCPDataRate);
	}
	inline u_int32_t getPreambleLength() { return(PreambleLength); }
	inline double getPLCPDataRate() { return(PLCPDataRate); }
	
	inline u_int32_t getPLCPhdrLen() {
		return((PreambleLength + PLCPHeaderLength) >> 3);
	}

	inline u_int32_t getHdrLen11() {
		return(getPLCPhdrLen() + sizeof(struct hdr_mac802_11)
                       + ETHER_FCS_LEN);
	}
	
	inline u_int32_t getRTSlen() {
		return(getPLCPhdrLen() + sizeof(struct rts_frame));
	}
	
	inline u_int32_t getCTSlen() {
		return(getPLCPhdrLen() + sizeof(struct cts_frame));
	}
	
	inline u_int32_t getACKlen() {
		return(getPLCPhdrLen() + sizeof(struct ack_frame));
	}

	u_int32_t getBeaconLen(struct hdr_beacon *bh);

 private:
	u_int32_t	CWMin;
	u_int32_t	CWMax;
	double		SlotTime;
	double 		CCATime;
	double		RxTxTurnaroundTime;
	double		SIFSTime;
	u_int32_t	PreambleLength;
	u_int32_t	PLCPHeaderLength;
	double		PLCPDataRate;
};


/*
 * IEEE 802.11 Spec, section 11.4.4.2
 *      - default values for the MAC Attributes
 */

#define MAC_RTSThreshold		3000	// bytes
#define MAC_ATIMRetryLimit		3	// ATIM retries
#define MAC_RTSRetryLimit		10	// RTS retry limits
#define MAC_ShortRetryLimit		7	// retransmittions
#define MAC_LongRetryLimit		4	// retransmissions
#define MAC_FragmentationThreshold	2346	// bytes
#define MAC_MaxTransmitMSDULifetime	512	// time units
#define MAC_MaxReceiveLifetime		512	// time units
const double MAC_TBTTPreWake=           0.00001;  // 10 us

class MAC_MIB {
public:

	MAC_MIB(Mac802_11 *parent);

private:
	u_int32_t	RTSThreshold;
	u_int32_t	ShortRetryLimit;
	u_int32_t	LongRetryLimit;
//	u_int32_t	FragmentationThreshold;
//	u_int32_t	MaxTransmitMSDULifetime;
//	u_int32_t	MaxReceiveLifetime;
//	u_int32_t	TransmittedFragmentCount;
//	u_int32_t	MulticastTransmittedFrameCount;
	u_int32_t	RetryCount;
	u_int32_t	MultipleRetryCount;
	u_int32_t	FrameDuplicateCount;
	u_int32_t	RTSSuccessCount;
	u_int32_t	ReceivedFragmentCount;
	u_int32_t	MulticastReceivedFrameCount;
	u_int32_t	FCSErrorCount;

public:
	u_int32_t	FailedCount;	
	u_int32_t	RTSFailureCount;
	u_int32_t	ACKFailureCount;

        double          BeaconPeriod;
        double          ATIMWindow;
        double          TBTTPreWake;
	double		AdvWindow;

public:
       inline u_int32_t getRTSThreshold() { return(RTSThreshold);}
       inline u_int32_t getShortRetryLimit() { return(ShortRetryLimit);}
       inline u_int32_t getLongRetryLimit() { return(LongRetryLimit);}
};


/* ======================================================================
   The following destination class is used for duplicate detection.
   ====================================================================== */
class Host {
public:
	LIST_ENTRY(Host) link;
	u_int32_t	index;
	u_int32_t	seqno;
};

class buffer_entry;

/* ======================================================================
   The actual 802.11 MAC class.
   ====================================================================== */
class Mac802_11 : public Mac {
	friend class DeferTimer;
	friend class BackoffTimer;
	friend class BackoffInterruptTimer;
	friend class IFTimer;
	friend class NavTimer;
	friend class RxTimer;
	friend class TxTimer;
	friend class TBTTTimer;
	friend class BeaconTimer;
	friend class AdvEndTimer;
	friend class ATIMEndTimer;
	friend class buffer_entry;
	friend class God;

public:
	Mac802_11();
	void		recv(Packet *p, Handler *h);
	int	hdr_dst(char* hdr, int dst = -2);
	int	hdr_src(char* hdr, int src = -2);
	inline int	hdr_type(char* hdr, u_int16_t type = 0);
	
	inline int bss_id() { return bss_id_; }
	
	// Added by Sushmita to support event tracing
        void trace_event(char *, Packet *);
        EventTrace *et_;

        // Added for PSM support
	double	bftime();
	int i_am_idle();
	void handleBackoffInterrupt();

	// scan and reclaim packets
	Packet * reclaim_packet(int dst);

protected:
	void	backoffHandler(void);
	void	deferHandler(void);
	void	navHandler(void);
	void	recvHandler(void);
	void	sendHandler(void);
	void	txHandler(void);
	void    tbttHandler(void);
	void    beaconHandler(void);
	void    advEndHandler(void);
	void    atimEndHandler(void);

private:
	int	command(int argc, const char*const* argv);
	void	Terminate(void); /* called at end of simulation */
	void	callback(); // to sendup callback_

	/*
	 * Called by the timers.
	 */
	void		recv_timer(void);
	void		send_timer(void);
	int		check_pktCTRL();
	int		check_pktRTS();
	int		check_pktTx();
	int             check_pktBEACON();
	int             check_pktATIM();

	/*
	 * Packet Transmission Functions.
	 */
	void	send(Packet *p, Handler *h);
	void 	sendRTS(int dst);
	void	sendCTS(int dst, double duration);
	void	sendACK(int dst);
	void	sendDATA(Packet *p);
	void	sendBEACON();
	void	RetransmitRTS();
	void	RetransmitDATA();
	void	RetransmitATIM();

	/*
	 * Packet Reception Functions.
	 */
	void	recvRTS(Packet *p);
	void	recvCTS(Packet *p);
	void	recvACK(Packet *p);
	void	recvDATA(Packet *p);
	void	recvBEACON(Packet *p);
	void	recvATIM(Packet *p);


	// these two functions scan the buffer and copy the appropriate
	// packets into pktTx/pktRTS/pktATIM.
	void		atim_scan();
	bool		tx_scan();


	void		capture(Packet *p);
	void		collision(Packet *p);
	void		discard(Packet *p, const char* why);
	void		rx_resume(void);
	void		tx_resume(void);
	void		goto_sleep();

	int	is_idle(void);
	
	/*
	 * Debugging Functions.
	 */
	void		trace_pkt(Packet *p);
	void		dump(char* fname);

	// int		initialized_;
	inline int initialized() {	
		return (cache_ && logtarget_
                        && Mac::initialized());
	}

	inline void mac_log(Packet *p) {
                logtarget_->recv(p, (Handler*) 0);
        }

	double txtime(Packet *p);
	double txtime(double psz, double drt);
	double txtime(int bytes) { /* clobber inherited txtime() */ abort(); return 0;}

	inline void transmit(Packet *p, double timeout);
	inline void checkBackoffTimer(void);
	inline void postBackoff(int pri);
	inline void setRxState(MacState newState);
	inline void setTxState(MacState newState);

	inline void inc_cw() {
		cw_ = (cw_ << 1) + 1;
		if(cw_ > phymib_.getCWMax())
			cw_ = phymib_.getCWMax();
	}
	inline void rst_cw() { cw_ = phymib_.getCWMin(); }

	inline double sec(double t) { return(t *= 1.0e-6); }
	inline u_int16_t usec(double t) {
		u_int16_t us = (u_int16_t)floor((t *= 1e6) + 0.5);
		return us;
	}
	inline void set_nav(u_int16_t us) {
		double now = Scheduler::instance().clock();
		double t = us * 1e-6;
		if((now + t) > nav_) {
			nav_ = now + t;
			if(mhNav_.busy())
				mhNav_.stop();
			mhNav_.start(t);
		}
	}

	// PSM statistics variables
	int total_retransmit;
	double total_wasted_time;
	double idle_start;
	double total_true_idle_time;

protected:
	PHY_MIB         phymib_;
        MAC_MIB         macmib_;

       /* the macaddr of my AP in BSS mode; for IBSS mode
        * this is set to a reserved value IBSS_ID - the
        * MAC_BROADCAST reserved value can be used for this
        * purpose
        */
       int     bss_id_;
       enum    {IBSS_ID=MAC_BROADCAST};

private:
	double		basicRate_;
 	double		dataRate_;

        // (SPAN) assume that routers are eternally awake
        bool span_fastroute_;   

        // offset of beacon header
        int off_beacon_;

	// offset of span header
        int off_span_;

	// For Debugging reentrancy!
	int entry_;

        EnergyModel *energy_model_;
        MobileNode *node_;
        CIRCLEQ_HEAD(buffer_head, buffer_entry) buf_head_;
	        
	// have we {sent, received} an ATIM before?
	bool first_atim_, recvd_atim_;
	unsigned has_packet_tx_;
	unsigned recvd_bcast_atim_;
	unsigned recvd_ucast_atim_;
	unsigned initial_wake_count_;

	bool span_opt_;
	bool psm_mode_;
	bool got_beacon_; 

	// address we last sent an ATIM to
	u_char	last_atim_ra_[ETHER_ADDR_LEN]; 
	
	/*
	 * Mac Timers
	 */
	IFTimer		mhIF_;		// interface timer
	NavTimer	mhNav_;		// NAV timer
	RxTimer		mhRecv_;		// incoming packets
	TxTimer		mhSend_;		// outgoing packets

	DeferTimer	mhDefer_;	// defer timer
	BackoffTimer	mhBackoff_;	// backoff timer
        TBTTTimer       mhTBTT_;         // TBTT timer, schedule start of ATIM
        BeaconTimer     mhBeacon_;       // beacon timer
        AdvEndTimer     mhAdvEnd_;       // end of adv window
        ATIMEndTimer    mhEndATIM_;      // ATIM end of window
        TSFTimer        tsfTimer_; 

	/* ============================================================
	   Internal MAC State
	   ============================================================ */
	double		nav_;		// Network Allocation Vector

	MacState	rx_state_;	// incoming state (MAC_RECV or MAC_IDLE)
	MacState	tx_state_;	// outgoint state
	int		tx_active_;	// transmitter is ACTIVE

	Packet          *eotPacket_;    // copy for eot callback

	Packet		*pktRTS_;	// outgoing RTS packet
	Packet		*pktCTRL_;	// outgoing non-RTS packet
        Packet          *pktBEACON_;     // outgoing Beacon packet
        Packet          *pktATIM_;       // outgoing ATIM packet
        buffer_entry    *entryATIM_;     // outgoing ATIM packet's buf entry

	u_int32_t	cw_;		// Contention Window
	u_int32_t	ssrc_;		// STA Short Retry Count
	u_int32_t	slrc_;		// STA Long Retry Count
	u_int32_t	atimrc_;	// ATIM Retry Count
	u_int32_t	rtsrc_;		// RTS Retry Count

/*
	double		sifs;		// Short Interface Space
	double		pifs;		// PCF Interframe Space
	double		difs;		// DCF Interframe Space
	double		eifs;		// Extended Interframe Space
	double		tx_sifs;
	double		tx_pifs;
	double		tx_difs;
	int             tx_opt;         // Optimistic tx
*/
	int		min_frame_len_;

	NsObject*	logtarget_;
	NsObject*       EOTtarget_;     // given a copy of packet at TX end


	/* ============================================================
	   Duplicate Detection state
	   ============================================================ */
	u_int16_t	sta_seqno_;	// next seqno that I'll use
	int		cache_node_count_;
	Host		*cache_;
};


/*
 * An entry in the Packet buffer
 */
class buffer_entry {
public:
    buffer_entry(u_int32_t addr, bool sent_atim,
		   bool recvd_ack, bool sent_data, int age) :
	packet_(0), addr_(addr), sent_atim_(sent_atim),
	recvd_ack_(recvd_ack), sent_data_(sent_data), 
	pwr_mgt_(true), pwr_mgt_updated_(false), age_(age) {}

    buffer_entry(Packet *packet, bool sent_atim,
		   bool recvd_ack, bool sent_data, int age) :
	packet_(packet), addr_(0), sent_atim_(sent_atim),
	recvd_ack_(recvd_ack), sent_data_(sent_data), 
	pwr_mgt_(true), pwr_mgt_updated_(false), age_(age) {}
    
    Packet *packet_;
    u_int32_t addr_;
    bool sent_atim_;
    bool recvd_ack_;
    bool sent_data_;
    bool pwr_mgt_;
    bool pwr_mgt_updated_;
    unsigned int age_;

    CIRCLEQ_ENTRY(buffer_entry) entries;

    u_int32_t addr(Mac802_11 *mac) {
	if (packet_) {
	    struct hdr_mac802_11 *dh = 
		(struct hdr_mac802_11 *)hdr_mac::access(packet_);
	    return ETHER_ADDR(dh->dh_ra);
	}
	return addr_;
    }

    void dump(Mac802_11 *mac) {
	int uid = -1;
	if (packet_) {
	    struct hdr_cmn *ch = HDR_CMN(packet_);
	    uid = ch->uid_;
	}
#if 0
	fprintf(stdout, "%f: buffer entry to %x: sent ATIM=%d, received ACK=%d, sent DATA=%d, age=%u, uid=%d\n",
		Scheduler::instance().clock(),
		addr(mac), 
		sent_atim_, 
		recvd_ack_, 
		sent_data_, 
		age_,
		uid);
#endif
#ifdef DEBUG
	// let's check that this atim has a state that we expect!
	#define sa sent_atim_ 
	#define ra recvd_ack_
	#define sd sent_data_
	assert( (!sa && !ra && !sd) ||
		(sa && !ra && !sd) ||
		(!sa && ra && !sd) ||
		(sa && ra && !sd) ||
		
		(sa && ra && sd) ||
		(!sa && ra && sd) ||
		(sa && !ra && sd));
#endif
    }
};

/*
 * packet buffer entry states:
 * 
 * sent_atim     recvd_ack     sent_data     meaning
 * 0             0             0
 * 1             0             0
 * 0             1             0
 * 1             1             0
 * ----------------------------------------------------------------------------
 * 1             1             1
 * 0             1             1
 * 1             0             1
 *
 */


#endif /* __mac_80211_h__ */

